"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __param = (this && this.__param) || function (paramIndex, decorator) {
    return function (target, key) { decorator(target, key, paramIndex); }
};
Object.defineProperty(exports, "__esModule", { value: true });
const inversify_1 = require("inversify");
const ModulesRegistry_1 = require("../modules/ModulesRegistry");
const Cli_1 = require("../cli/Cli");
const ManifestsScanner_1 = require("../manifests/ManifestsScanner");
const Exception_1 = require("../error/Exception");
const ModelsBuilder_1 = require("../../domain/ModelsBuilder");
const Models = require("../../domain/Models");
const Path = require("path");
const _ = require("lodash");
const fs = require("fs");
const OpsApplicationConfiguration_1 = require("../../application/OpsApplicationConfiguration");
const inquirer = require("inquirer");
const YamlService_1 = require("../yaml/YamlService");
const traverse = require("traverse");
const YamlLazyValue_1 = require("../yaml/YamlLazyValue");
const semver = require("semver");
const LazyModule_1 = require("../../domain/models/LazyModule");
const packageJson = require("./../../../package.json");
let ModulesService = class ModulesService {
    constructor($applicationConfiguration, $cli, $manifestsScanner, $modelsBuilder, $yaml) {
        this.$applicationConfiguration = $applicationConfiguration;
        this.$cli = $cli;
        this.$manifestsScanner = $manifestsScanner;
        this.$modelsBuilder = $modelsBuilder;
        this.$yaml = $yaml;
        this.modulesRoot = "";
        this.registries = new Map();
        this.cache = new Map();
    }
    setModulesRoot(modulesRoot) {
        this.modulesRoot = modulesRoot;
    }
    getRegistry(id = "default") {
        if (!this.hasRegistry(id)) {
            throw new Exception_1.default("Registry is not found", { RegistryId: id });
        }
        return this.registries.get(id);
    }
    hasRegistry(id) {
        return this.registries.has(id);
    }
    readModule(manifestPath, context = {}) {
        if (this.cache.has(manifestPath)) {
            this.$cli.logTrace(`Reading module manifest '${manifestPath}' (cached)`);
            return this.cache.get(manifestPath);
        }
        this.$cli.logTrace(`Reading module manifest '${manifestPath}'`);
        const modulePath = Path.dirname(Path.relative(this.modulesRoot, manifestPath));
        const moduleId = modulePath == "." ? "." : "./" + modulePath;
        const content = this.$yaml.parseFile(manifestPath);
        const manifest = _.merge({}, content, {
            module: {
                id: moduleId,
                paths: {
                    module: moduleId,
                    manifest: manifestPath.toString(),
                    directory: Path.dirname(manifestPath.toString()),
                    unit: Path.basename(Path.dirname(manifestPath.toString())),
                    root: fs.realpathSync(this.modulesRoot)
                }
            }
        }, context);
        try {
            const module = this.$modelsBuilder.buildModule(manifest);
            if (!semver.satisfies(packageJson.version, module.metadata.schema)) {
                throw new Exception_1.default("Module schema not supported by current ops version!", {
                    ModuleId: moduleId,
                    File: manifestPath,
                    Schema: module.metadata.schema,
                    Version: packageJson.version,
                    Source: fs.readFileSync(manifestPath).toString()
                });
            }
            return module;
        }
        catch (error) {
            throw new Exception_1.default("Unable to build module model from manifest!", {
                ModuleId: moduleId,
                File: manifestPath,
                Source: fs.readFileSync(manifestPath).toString()
            }, error);
        }
    }
    readInterpolatedModule(manifestPath, context = {}) {
        if (this.cache.has(manifestPath)) {
            this.$cli.logTrace(`Reading interpolated module manifest '${manifestPath}' (cached)`);
            return this.cache.get(manifestPath);
        }
        this.$cli.logTrace(`Reading interpolated module manifest '${manifestPath}'`);
        const modulePath = Path.dirname(Path.relative(this.modulesRoot, manifestPath));
        const moduleId = modulePath == "." ? "." : "./" + modulePath;
        const content = this.$yaml.parseJsonFile(manifestPath);
        const manifest = _.merge({}, content, {
            module: {
                id: moduleId,
                paths: {
                    module: moduleId,
                    manifest: manifestPath.toString(),
                    directory: Path.dirname(manifestPath.toString()),
                    unit: Path.basename(Path.dirname(manifestPath.toString())),
                    root: fs.realpathSync(this.modulesRoot)
                }
            }
        }, context);
        try {
            const module = this.$modelsBuilder.buildModule(manifest);
            if (!semver.satisfies(packageJson.version, module.metadata.schema)) {
                throw new Exception_1.default("Module schema not supported by current ops version!", {
                    ModuleId: moduleId,
                    File: manifestPath,
                    Schema: module.metadata.schema,
                    Version: packageJson.version,
                    Source: fs.readFileSync(manifestPath).toString()
                });
            }
            this.cache.set(manifestPath, module);
            return module;
        }
        catch (error) {
            throw new Exception_1.default("Unable to build module model from manifest!", {
                ModuleId: moduleId,
                File: manifestPath,
                Source: fs.readFileSync(manifestPath).toString()
            }, error);
        }
    }
    createRegistry(id, context = {}) {
        this.$cli.logTrace(`Initializing '${id}' modules registry:`);
        const registry = new ModulesRegistry_1.default(context);
        const manifestPaths = this.scanManifestPaths();
        const modules = manifestPaths.map((manifestPath) => {
            const modulePath = Path.dirname(Path.relative(this.modulesRoot, manifestPath));
            const moduleId = modulePath == "." ? "." : "./" + modulePath;
            return new LazyModule_1.default(moduleId, this.readModule.bind(this, manifestPath, context));
        });
        registry.addLazyModules(modules);
        this.$cli.logTrace(`Loaded ${registry.size} modules!`);
        this.registries.set(id, registry);
        return registry;
    }
    queryModules(selector, registryId) {
        if (typeof selector == "string")
            selector = { pattern: [selector] };
        const registry = this.getRegistry(registryId);
        const modules = registry.getModules();
        const lazyModules = registry.getLazyModules();
        this.$cli.logTrace("Querying modules:", selector, "Registry", registryId);
        return _.uniq([...modules, ...lazyModules]
            .filter((target) => {
            let filtered = false;
            selector = selector;
            if (selector.filter) {
                filtered = true;
                if (!selector.filter(target, selector))
                    return false;
            }
            if (selector.id) {
                filtered = true;
                const selectorIds = selector.id instanceof Array ?
                    selector.id :
                    [selector.id];
                if (!selectorIds.includes(target.id))
                    return false;
            }
            if (selector.tag && target instanceof Models.Module) {
                filtered = true;
                const selectorTags = selector.tag instanceof Array ?
                    selector.tag :
                    [selector.tag];
                const matches = selectorTags.filter((selectorTag) => target.metadata.tags.has(selectorTag));
                // Expect all tags
                if (matches.length != selectorTags.length)
                    return false;
            }
            if (selector.ancestorOf) {
                filtered = true;
                const selectorAncestors = selector.ancestorOf instanceof Array ?
                    selector.ancestorOf :
                    [selector.ancestorOf];
                const matches = selectorAncestors.filter((selectorAncestor) => {
                    let selectorAncestorIdChunks = selectorAncestor.split("/");
                    while (selectorAncestorIdChunks.length) {
                        selectorAncestorIdChunks = selectorAncestorIdChunks.slice(0, selectorAncestorIdChunks.length - 1);
                        if (target.id == selectorAncestorIdChunks.join("/"))
                            return true;
                    }
                    return false;
                });
                if (!matches.length)
                    return false;
            }
            if (selector.parentOf) {
                filtered = true;
                const selectorParents = selector.parentOf instanceof Array ?
                    selector.parentOf :
                    [selector.parentOf];
                const matches = selectorParents.filter((selectorParent) => {
                    const selectorParentIdChunks = selectorParent.split("/");
                    const selectorParentId = selectorParentIdChunks.slice(0, selectorParentIdChunks.length - 1).join("/");
                    return target.id == selectorParentId;
                });
                if (!matches.length)
                    return false;
            }
            if (selector.pattern) {
                filtered = true;
                const selectorPatterns = selector.pattern instanceof Array ?
                    selector.pattern :
                    [selector.pattern];
                const matches = selectorPatterns.filter((selectorPattern) => target.id.match(new RegExp("^" + selectorPattern
                    .replace(/\./g, "\.")
                    .replace(/\//g, "\/")
                    .replace(/[*]{2}/g, ".+")
                    .replace(/[*]/g, "[a-zA-Z0-9_-]+") + "$", "g")));
                if (!matches.length)
                    return false;
            }
            if (!filtered) {
                throw new Exception_1.default("Invalid selector", selector);
            }
            return true;
        })
            .map((module) => module.id)
            .sort((a, b) => {
            if (a == ".")
                return -1;
            if (b == ".")
                return 1;
            return a.split("/") < b.split("/") ? -1 : 1;
        }));
    }
    hasModule(moduleId, registryId) {
        const registry = this.getRegistry(registryId);
        return registry.hasModule(moduleId) || registry.hasLazyModule(moduleId);
    }
    queryModule(selector, registry) {
        return this.queryModules(selector, registry)[0];
    }
    async getModule(moduleId, registryId) {
        this.$cli.logTrace(`Getting module '${moduleId}' (${registryId})...`);
        const module = this.lookupModule(moduleId, registryId);
        if (module.metadata.interpolation == "completed") {
            return module;
        }
        if (module.metadata.interpolation != "none") {
            throw new Error(`Module '${module.id}' interpolation in progress!`);
        }
        this.$cli.logTrace(`Loading module '${moduleId}' (${registryId})...`);
        module.metadata.interpolation = "in-progress";
        if (module.id != ".") {
            let ancestorId = this
                .queryModules({ ancestorOf: module.id }, registryId)
                .reverse()[0];
            if (this.hasModule(ancestorId, registryId)) {
                let ancestor = this.lookupModule(ancestorId, registryId);
                switch (ancestor.metadata.interpolation) {
                    case "completed":
                        if (!module.metadata.parents.has(ancestor.id)) {
                            this.$cli.logTrace(`Inheriting ancestor '${ancestor.id}' –> '${module.id}'...`);
                            module.inheritFrom(ancestor);
                        }
                        break;
                    case "none":
                        ancestor = await this.getModule(ancestorId, registryId);
                        if (!module.metadata.parents.has(ancestor.id)) {
                            this.$cli.logTrace(`Inheriting ancestor '${ancestor.id}' –> '${module.id}'...`);
                            module.inheritFrom(ancestor);
                        }
                        break;
                    default:
                        this.$cli.logTrace(`Inheriting ancestor '${ancestor.id}' –> '${module.id}' is pending...`);
                        break;
                }
            }
        }
        //
        const providers = this.queryModules({
            filter: (target) => {
                if (target instanceof Models.Module) {
                    if (!target.metadata.inherit.into.selector.tag) {
                        return false;
                    }
                    const selectorTags = target.metadata.inherit.into.selector.tag instanceof Array ?
                        target.metadata.inherit.into.selector.tag :
                        [target.metadata.inherit.into.selector.tag];
                    const matches = selectorTags.filter((selectorTag) => module.metadata.tags.has(selectorTag));
                    return matches.length == selectorTags.length;
                }
                else {
                    return false;
                }
            }
        }, registryId);
        //
        if (providers.length) {
            for (let providerIndex = 0; providerIndex < providers.length; providerIndex++) {
                const providerId = providers[providerIndex];
                let provider = this.lookupModule(providerId, registryId);
                switch (provider.metadata.interpolation) {
                    case "completed":
                        if (!module.metadata.parents.has(provider.id)) {
                            this.$cli.logTrace(`Inheriting provider '${provider.id}' –> '${module.id}'...`);
                            module.inheritFrom(provider);
                        }
                        break;
                    case "none":
                        provider = await this.getModule(providerId, registryId);
                        if (!module.metadata.parents.has(provider.id)) {
                            this.$cli.logTrace(`Inheriting provider '${provider.id}' –> '${module.id}'...`);
                            module.inheritFrom(provider);
                        }
                        break;
                    default:
                        break;
                }
            }
        }
        //
        if (!module.metadata.virtual) {
            const moduleInputs = module.inputs.entries;
            for (let moduleInputIndex = 0; moduleInputIndex < moduleInputs.length; moduleInputIndex++) {
                const moduleInput = moduleInputs[moduleInputIndex][1];
                this.$cli.logDebug((moduleInput.required ? "Required" : "Optional") + " input: ", moduleInput.name);
                if (this.$applicationConfiguration.context.inputs.hasOwnProperty(moduleInput.name)) {
                    moduleInput.value = this.$applicationConfiguration.context.inputs[moduleInput.name];
                    moduleInput.provided = true;
                    this.$cli.logTrace("Value provided from ops command line arguments:", moduleInput.value);
                }
                if (!moduleInput.provided && moduleInput.prompt) {
                    moduleInput.value = (await inquirer.prompt([
                        {
                            name: "input",
                            message: moduleInput.prompt.message,
                            type: moduleInput.prompt.type,
                            choices: (typeof moduleInput.prompt.choices != 'function' ?
                                moduleInput.prompt.choices :
                                (function () {
                                    return moduleInput.prompt.choices(module.serialize());
                                }())),
                            default: (typeof moduleInput.prompt.defaults != 'function' ?
                                moduleInput.prompt.defaults :
                                (function () {
                                    return moduleInput.prompt.defaults(module.serialize());
                                }()))
                        }
                    ])).input;
                    this.$applicationConfiguration.context.inputs[moduleInput.name] = moduleInput.value;
                    moduleInput.provided = true;
                }
                if (!moduleInput.provided && moduleInput.required) {
                    this.$cli.logError("Required input parameter", moduleInput.name, "is not provided!");
                    process.exit(1);
                }
            }
        }
        const cwd = module.paths.directory;
        const registry = this.getRegistry(registryId);
        const manifestObject = _.merge({
            module: module.serialize(),
            application: {
                env: process.env,
                context: this.$applicationConfiguration.context,
                version: this.$applicationConfiguration.version,
                requiredCommands: this.$applicationConfiguration.requiredCommands
            }
        }, registry.context);
        if (module.metadata.virtual) {
            traverse(manifestObject.module).forEach(function (element) {
                if (element instanceof YamlLazyValue_1.default) {
                    this.update("!lazy " + element.value);
                }
            });
        }
        else {
            traverse(manifestObject.module).forEach(function (element) {
                if (element instanceof YamlLazyValue_1.default) {
                    this.update(element.value);
                }
            });
        }
        let manifestYaml = this.$yaml.stringify(manifestObject);
        if (!module.metadata.virtual) {
            manifestYaml = manifestYaml
                .replace(/"<<-": '!lazy /g, '<<: \'')
                .replace(/<<-: '!lazy /g, '<<: \'')
                .replace(/"<<-": "!lazy /g, '<<: \"')
                .replace(/<<-: "!lazy /g, '<<: \"')
                .replace(/: '!lazy /g, ': \'')
                .replace(/: "!lazy /g, ': \"')
                .replace(/- '!lazy /g, '- \'')
                .replace(/- "!lazy /g, '- \"')
                .replace(/"<<-": !lazy /g, '<<: ')
                .replace(/<<-: !lazy /g, '<<: ')
                .replace(/: !lazy /g, ': ')
                .replace(/- !lazy /g, '- ')
                .replace(/"<<-": /g, "<<: ")
                .replace(/<<-: /g, "<<: ")
                .replace(/\s"!lazy /g, ' "');
        }
        if (providers.length) {
            manifestYaml = manifestYaml
                .replace(/<<-: '!direct-inheritor /g, '<<: \'')
                .replace(/<<-: "!direct-inheritor /g, '<<: \"')
                .replace(/: '!direct-inheritor /g, ': \'')
                .replace(/: "!direct-inheritor /g, ': \"')
                .replace(/- '!direct-inheritor /g, '- \'')
                .replace(/- "!direct-inheritor /g, '- \"')
                .replace(/\s"!direct-inheritor /g, ' "');
        }
        const rawFile = `${cwd}/~~module.manifest.raw.yaml`;
        const interpolatedFile = `${cwd}/~~module.manifest.interpolated.yaml`;
        try {
            fs.writeFileSync(rawFile, manifestYaml);
            const interpolatedYaml = this.$yaml.spiffFiles([rawFile], {
                asJson: true,
                env: module.env.filter((value) => !(value instanceof YamlLazyValue_1.default)).serialize()
            });
            fs.writeFileSync(interpolatedFile, interpolatedYaml);
            const interpolated = this.readInterpolatedModule(interpolatedFile, {
                module: {
                    metadata: {
                        $interpolation: "completed",
                        $parents: module.metadata.parents
                    }
                }
            });
            registry.setModule(module.id, interpolated);
            this.$cli.logDebug(`Loaded module '${module.id}'...`);
            return interpolated;
        }
        catch (error) {
            throw new Exception_1.default(`Unable to interpolate module!`, {
                ModuleId: module.id
            }, error);
        }
    }
    scanManifestPaths() {
        this.$cli.logTrace(`Scanning module in '${this.modulesRoot}'...`);
        return this.$manifestsScanner.find(this.modulesRoot);
    }
    lookupModule(moduleId, registryId) {
        const registry = this.getRegistry(registryId);
        if (registry.hasLazyModule(moduleId)) {
            this.$cli.logTrace(`Loading lazy module '${moduleId}' (${registryId})...`);
            const lazyModule = registry.getLazyModule(moduleId);
            const module = lazyModule.read();
            registry.setModule(moduleId, module);
            if (module.metadata.autoload.size) {
                const autoloadModuleIds = this.queryModules({
                    pattern: Array.from(module.metadata.autoload)
                }, registryId);
                autoloadModuleIds.forEach((moduleId) => {
                    this.$cli.logTrace(`Auto-loading module '${moduleId}' (${registryId})...`);
                    this.lookupModule(moduleId, registryId);
                });
            }
            return module;
        }
        else if (registry.hasModule(moduleId)) {
            return registry.getModule(moduleId);
        }
        else {
            throw new Exception_1.default("Module is not found", {
                ModuleId: moduleId,
                RegistryId: registryId
            });
        }
    }
};
ModulesService = __decorate([
    inversify_1.injectable(),
    __param(0, inversify_1.inject(OpsApplicationConfiguration_1.default)),
    __param(1, inversify_1.inject(Cli_1.default)),
    __param(2, inversify_1.inject(ManifestsScanner_1.default)),
    __param(3, inversify_1.inject(ModelsBuilder_1.default)),
    __param(4, inversify_1.inject(YamlService_1.default)),
    __metadata("design:paramtypes", [OpsApplicationConfiguration_1.default,
        Cli_1.default,
        ManifestsScanner_1.default,
        ModelsBuilder_1.default,
        YamlService_1.default])
], ModulesService);
exports.default = ModulesService;
//# sourceMappingURL=ModulesService.js.map