import "reflect-metadata";
import {inject, injectable} from "inversify";
import Cli from "../components/cli/Cli";
import Exception from "../components/error/Exception";
import * as tmp from "tmp";
import InterpolateApplicationConfiguration from "./InterpolateApplicationConfiguration";
import InterpolateApplicationUI from "./InterpolateApplicationUI";
import InterpolateApplicationContextReader from "./InterpolateApplicationContextReader";
import InterpolateApplicationContext from "./InterpolateApplicationContext";
import YamlService from "../components/yaml/YamlService";
import * as fs from "fs";
import * as Path from "path";
import * as traverse from "traverse";
import YamlLazyValue from "../components/yaml/YamlLazyValue";

@injectable()
export default class InterpolateApplication {

    constructor(
        @inject(Cli)
        private readonly $cli: Cli,
        @inject(InterpolateApplicationConfiguration)
        private readonly $configuration: InterpolateApplicationConfiguration,
        @inject(InterpolateApplicationUI)
        private readonly $ui: InterpolateApplicationUI,
        @inject(InterpolateApplicationContextReader)
        private readonly $contextReader: InterpolateApplicationContextReader,
        @inject(YamlService)
        private readonly $yaml: YamlService
    ) {
    }

    async run() {
        this.validateEnvironment();
        this.loadApplicationContext();
        this.interpolate();
    }

    private validateEnvironment() {
        const commandExists = require('command-exists').sync;
        this.$configuration.requiredCommands.map((command: string) => {
            if (!commandExists(command)) {
                throw new Exception(
                    `Missing required command!\nPlease see README for the details!`,
                    {Command: command}
                );
            }
        });
    }

    private validateContext(context: InterpolateApplicationContext) {
        if (!context.inputFile) {
            throw "Invalid input file!";
        }
    }

    private loadApplicationContext() {
        const context = this.$contextReader.readContext();
        if (context.quiet) this.$cli.quiet = true;
        switch (context.logLevel) {
            case "error":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = false;
                this.$cli.levels.info = false;
                this.$cli.levels.debug = false;
                this.$cli.levels.trace = false;
                break;
            case "warn":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = true;
                this.$cli.levels.info = false;
                this.$cli.levels.debug = false;
                this.$cli.levels.trace = false;
                break;
            case "info":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = true;
                this.$cli.levels.info = true;
                this.$cli.levels.debug = false;
                this.$cli.levels.trace = false;
                break;
            case "debug":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = true;
                this.$cli.levels.info = true;
                this.$cli.levels.debug = true;
                this.$cli.levels.trace = false;
                break;
            case "trace":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = true;
                this.$cli.levels.info = true;
                this.$cli.levels.debug = true;
                this.$cli.levels.trace = true;
                break;
        }
        if (context.help) {
            this.$ui.showApplicationHelp();
            process.exit(0);
        } else {
            this.$ui.showApplicationInfo(context);
            try {
                this.validateContext(context);
            } catch (error) {
                this.$cli.logError(error);
                this.$ui.showApplicationHelp();
                process.exit(1);
            }
        }
        this.$configuration.context.tempDir = context.tempDir;
        this.$configuration.context.logLevel = context.logLevel;
        this.$configuration.context.inputFile = context.inputFile;
        this.$configuration.context.output = context.output;
        this.$configuration.context.help = context.help;
        this.$configuration.context.quiet = context.quiet;
        this.$configuration.context.mergeFiles = context.mergeFiles;
        this.$configuration.context.contextFile = context.contextFile;
        if (!this.$configuration.context.tempDir) {
            tmp.setGracefulCleanup();
            this.$configuration.context.tempDir = tmp.dirSync({unsafeCleanup: true}).name;
        }
    }

    private interpolate() {
        const context = this.$configuration.context;
        const pathInput = context.inputFile;
        const mergeFiles = context.mergeFiles || [];
        const contextFile = context.contextFile;
        const cwd = Path.dirname(pathInput);
        let inputContent = this.$yaml.parseFile(
            pathInput,
            {
                context: contextFile ? this.$yaml.parseFile(contextFile) : undefined
            }
        );
        traverse(inputContent).forEach(function (element) {
            if (element instanceof YamlLazyValue) {
                this.update("!lazy " + element.value);
            }
        });
        let source = this.$yaml.stringify(inputContent, {path: pathInput})
            .replace(/"<<-": '!lazy /g, '<<: \'')
            .replace(/<<-: '!lazy /g, '<<: \'')
            .replace(/"<<-": "!lazy /g, '<<: \"')
            .replace(/<<-: "!lazy /g, '<<: \"')
            .replace(/: '!lazy /g, ': \'')
            .replace(/: "!lazy /g, ': \"')
            .replace(/- '!lazy /g, '- \'')
            .replace(/- "!lazy /g, '- \"')
            .replace(/"<<-": !lazy /g, '<<: ')
            .replace(/<<-: !lazy /g, '<<: ')
            .replace(/: !lazy /g, ': ')
            .replace(/- !lazy /g, '- ')
            .replace(/"<<-": /g, "<<: ")
            .replace(/<<-: /g, "<<: ");
        const pathInterpolated = cwd + "/~~" + Path.basename(pathInput) + ".interpolate.yaml";
        fs.writeFileSync(pathInterpolated, source);
        source = this.$yaml.spiffFiles([pathInterpolated, ...mergeFiles], {
            cwd: cwd
        });
        fs.unlinkSync(pathInterpolated);
        if (context.output) {
            fs.writeFileSync(context.output, source);
        } else {
            console.log(source);
        }
    }
}
