import "reflect-metadata";
import {inject, injectable} from "inversify";
import OpsApplicationContextReader from "./OpsApplicationContextReader";
import Cli from "../components/cli/Cli";
import OpsApplicationUI from "./OpsApplicationUI";
import * as fs from "fs";
import * as _ from "lodash";
import OpsApplicationContext from "./OpsApplicationContext";
import OpsApplicationConfiguration from "./OpsApplicationConfiguration";
import TaskExecutionService from "../components/tasks/TaskExecutionService";
import TaskExecutionError from "../components/tasks/TaskExecutionError";
import EchoTaskExecutor from "../components/tasks/executors/echo/EchoTaskExecutor";
import TerraformSourcesFormatterTaskExecutor
    from "../components/tasks/executors/terraform/TerraformSourcesFormatterTaskExecutor";
import TerragruntConfigurationGeneratorTaskExecutor
    from "../components/tasks/executors/terraform/TerragruntConfigurationGeneratorTaskExecutor";
import TerraformTaskExecutor from "../components/tasks/executors/terraform/TerraformTaskExecutor";
import ManifestTaskExecutor from "../components/tasks/executors/manifest/ManifestTaskExecutor";
import TerraformSourcesImporterTaskExecutor
    from "../components/tasks/executors/terraform/TerraformSourcesImporterTaskExecutor";
import OpsTaskExecutor from "../components/tasks/executors/ops/OpsTaskExecutor";
import TerraformVariablesGeneratorTaskExecutor
    from "../components/tasks/executors/terraform/TerraformVariablesGeneratorTaskExecutor";
import FSTaskExecutor from "../components/tasks/executors/fs/FSTaskExecutor";
import Exception from "../components/error/Exception";
import * as Models from "../domain/Models";
import ModulesService from "../components/modules/ModulesService";
import TerraformSectionSourcesGeneratorTaskExecutor
    from "../components/tasks/executors/terraform/TerraformSectionSourcesGeneratorTaskExecutor";
import TerraformTemplateSourcesGeneratorTaskExecutor
    from "../components/tasks/executors/terraform/TerraformTemplateSourcesGeneratorTaskExecutor";
import TerraformRemoteStatesGeneratorTaskExecutor
    from "../components/tasks/executors/terraform/TerraformRemoteStatesGeneratorTaskExecutor";
import ExecTaskExecutor from "../components/tasks/executors/exec/ExecTaskExecutor";
import * as tmp from "tmp";
import * as Moment from "moment";

@injectable()
export default class OpsApplication {

    constructor(
        @inject(Cli)
        readonly $cli: Cli,
        @inject(OpsApplicationConfiguration)
        readonly $configuration: OpsApplicationConfiguration,
        @inject(OpsApplicationUI)
        private readonly $ui: OpsApplicationUI,
        @inject(OpsApplicationContextReader)
        private readonly $contextReader: OpsApplicationContextReader,
        @inject(ModulesService)
        private readonly $modulesService: ModulesService,
        @inject(TaskExecutionService)
        private readonly $taskExecutionService: TaskExecutionService,
        @inject(OpsTaskExecutor)
        private readonly $opsTaskExecutor: OpsTaskExecutor,
        @inject(FSTaskExecutor)
        private readonly $fsTaskExecutor: FSTaskExecutor,
        @inject(ManifestTaskExecutor)
        private readonly $manifestTaskExecutor: ManifestTaskExecutor,
        @inject(EchoTaskExecutor)
        private readonly $echoTaskExecutor: EchoTaskExecutor,
        @inject(ExecTaskExecutor)
        private readonly $execTaskExecutor: ExecTaskExecutor,
        @inject(TerraformVariablesGeneratorTaskExecutor)
        private readonly $terraformVariablesGeneratorTaskExecutor: TerraformVariablesGeneratorTaskExecutor,
        @inject(TerraformSectionSourcesGeneratorTaskExecutor)
        private readonly $terraformSectionSourcesGeneratorTaskExecutor: TerraformSectionSourcesGeneratorTaskExecutor,
        @inject(TerraformTemplateSourcesGeneratorTaskExecutor)
        private readonly $terraformTemplateSourcesGeneratorTaskExecutor: TerraformTemplateSourcesGeneratorTaskExecutor,
        @inject(TerraformSourcesImporterTaskExecutor)
        private readonly $terraformSourcesImporterTaskExecutor: TerraformSourcesImporterTaskExecutor,
        @inject(TerraformSourcesFormatterTaskExecutor)
        private readonly $terraformSourcesFormatterTaskExecutor: TerraformSourcesFormatterTaskExecutor,
        @inject(TerragruntConfigurationGeneratorTaskExecutor)
        private readonly $terragruntConfigurationGeneratorTaskExecutor: TerragruntConfigurationGeneratorTaskExecutor,
        @inject(TerraformTaskExecutor)
        private readonly $terraformTaskExecutor: TerraformTaskExecutor,
        @inject(TerraformRemoteStatesGeneratorTaskExecutor)
        private readonly $terraformRemoteStatesGeneratorTaskExecutor: TerraformRemoteStatesGeneratorTaskExecutor
    ) {
    }

    async run() {
        try {
            const time = +new Date();
            this.validateEnvironment();
            this.registerTaskExecutors();
            this.loadApplicationContext();
            await this.executeTarget();
            this.$cli.logHeader("Execution completed in", Moment.duration((+new Date) - time, "milliseconds").humanize(), "(" + ((+new Date) - time) + "ms)");
        } catch (error) {
            this.$cli.logError(error);
            this.$cli.logError(
                "To re-run application please execute:\n",
                _.flatten([
                    "ops",
                    this.$configuration.context.help ?
                        ["--help"] : [],
                    this.$configuration.context.targetModule ?
                        ["-m", this.$configuration.context.targetModule] : [],
                    this.$configuration.context.bindingModule ?
                        ["-b", this.$configuration.context.bindingModule] : [],
                    this.$configuration.context.targetLifecycle ?
                        ["-l", this.$configuration.context.targetLifecycle] : [],
                    this.$configuration.context.targetTask ?
                        ["-t", this.$configuration.context.targetTask] : [],
                    Object.keys(this.$configuration.context.inputs).length ?
                        ["-i", JSON.stringify(JSON.stringify(this.$configuration.context.inputs))] : [],
                    this.$configuration.context.logLevel ?
                        ["-v", this.$configuration.context.logLevel] : [],
                ]).join(" ")
            );
            process.exit(1);
        }
    }

    private validateEnvironment() {
        const commandExists = require('command-exists').sync;
        this.$configuration.requiredCommands.map((command: string) => {
            if (!commandExists(command)) {
                throw new Exception(
                    `Missing required command!\nPlease see README for the details!`,
                    {Command: command}
                );
            }
        });
    }

    private validateContext(context: OpsApplicationContext) {
        if (!context.targetModule.length)
            throw "Module name is not specified!";
        if (!context.targetTask && !context.targetLifecycle)
            throw "Lifecycle name or task name is not specified!";
        if (!context.modulesRoot.length)
            throw "Path to modules is not specified!";
        if (!fs.existsSync(context.modulesRoot))
            throw "Path to modules doesn't exists!";
        if (!fs.lstatSync(context.modulesRoot).isDirectory())
            throw "Path to modules is not directory!";
    }

    private loadApplicationContext() {
        const context = this.$contextReader.readContext();
        if (context.quiet) {
            this.$cli.quiet = true;
        }
        switch (context.logLevel) {
            case "error":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = false;
                this.$cli.levels.info = false;
                this.$cli.levels.debug = false;
                this.$cli.levels.trace = false;
                break;
            case "warn":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = true;
                this.$cli.levels.info = false;
                this.$cli.levels.debug = false;
                this.$cli.levels.trace = false;
                break;
            case "info":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = true;
                this.$cli.levels.info = true;
                this.$cli.levels.debug = false;
                this.$cli.levels.trace = false;
                break;
            case "debug":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = true;
                this.$cli.levels.info = true;
                this.$cli.levels.debug = true;
                this.$cli.levels.trace = false;
                break;
            case "trace":
                this.$cli.levels.error = true;
                this.$cli.levels.warn = true;
                this.$cli.levels.info = true;
                this.$cli.levels.debug = true;
                this.$cli.levels.trace = true;
                break;
        }
        if (context.help) {
            this.$ui.showApplicationHelp();
            process.exit(0);
        } else {
            this.$ui.showApplicationInfo(context);
            try {
                this.validateContext(context);
            } catch (error) {
                this.$cli.logError(error);
                this.$ui.showApplicationHelp();
                process.exit(1);
            }
        }
        this.$configuration.context.tempDir = context.tempDir;
        this.$configuration.context.logLevel = context.logLevel;
        this.$configuration.context.bindingModule = context.bindingModule;
        this.$configuration.context.modulesRoot = context.modulesRoot;
        this.$configuration.context.targetTask = context.targetTask;
        this.$configuration.context.targetModule = context.targetModule;
        this.$configuration.context.targetLifecycle = context.targetLifecycle;
        this.$configuration.context.help = context.help;
        this.$configuration.context.quiet = context.quiet;
        this.$configuration.context.inputs = context.inputs;
        this.$configuration.context.continueFrom = context.continueFrom;
        if (!this.$configuration.context.tempDir) {
            tmp.setGracefulCleanup();
            this.$configuration.context.tempDir = tmp.dirSync({unsafeCleanup: true}).name;
        }
        this.$modulesService.setModulesRoot(context.modulesRoot);
    }

    private async executeTarget() {
        const context = this.$configuration.context;
        const bindingModuleName = context.bindingModule;
        let bindingModule;
        if (bindingModuleName) {
            this.$cli.logTrace("Initializing registry for binding:", bindingModuleName);
            this.$modulesService.createRegistry("binding");
            if (!this.$modulesService.hasModule(bindingModuleName, "binding"))
                throw `Binding '${bindingModuleName}' is not found!`;
            bindingModule = await this.$modulesService.getModule(bindingModuleName, "binding");
            this.$modulesService.createRegistry("application", {
                binding: bindingModule.serializeAsBinding()
            });
        } else {
            this.$cli.logTrace("Initializing registry for application");
            this.$modulesService.createRegistry("application");
        }
        const targetModuleName = context.targetModule;
        const targetLifecycleName = context.targetLifecycle;
        const targetTask = context.targetTask;
        if (!this.$modulesService.hasModule(targetModuleName, "application"))
            throw `Module '${targetModuleName}' is not found!`;
        const targetModule = await this.$modulesService.getModule(targetModuleName, "application");
        if (targetLifecycleName && !targetModule.lifecycles.has(targetLifecycleName))
            throw `Module '${targetModuleName}' doesn't have lifecycle '${targetLifecycleName}'!`;
        if (targetTask && !targetModule.tasks.has(targetTask))
            throw `Module '${targetModuleName}' doesn't have task '${targetTask}'!`;
        if (!targetModule.metadata.enabled)
            throw `Module '${targetModuleName}' is not enabled!`;
        try {
            //this.$cli.logLifecycleStart(`Performing ${targetModuleName} ${targetLifecycleName ? `lifecycle:${targetLifecycleName}` : ""} ${targetTask ? `task:${targetTask}` : ""}`);
            const task = new Models.ModuleTask(
                "ops",
                new Models.ModuleTaskMetadata(
                    true,
                    "ops",
                    true,
                    "Internal ops task"
                ),
                {
                    selector: {
                        id: targetModuleName
                    },
                    execution: {
                        type: targetLifecycleName ? "lifecycle" : "task",
                        targets: targetLifecycleName ? [targetLifecycleName] : [targetTask],
                        ignoreErrors: context.ignoreErrors,
                        continueFrom: context.continueFrom
                    },
                    context: {
                        binding: bindingModule ? bindingModule.serializeAsBinding() : undefined
                    }
                }
            );
            await this.$taskExecutionService.executeTask(
                task,
                context,
                targetModule,
                bindingModule
            );
            this.$cli.logLifecycleEnd();
        } catch (error) {
            if (error instanceof TaskExecutionError) {
                this.$cli.logNamedYamlObject(error, `Module: ${targetModuleName}`);
            }
            throw error;
        }
    }

    private registerTaskExecutors() {
        [
            this.$opsTaskExecutor,
            this.$fsTaskExecutor,
            this.$manifestTaskExecutor,
            this.$echoTaskExecutor,
            this.$execTaskExecutor,
            this.$terraformVariablesGeneratorTaskExecutor,
            this.$terraformSectionSourcesGeneratorTaskExecutor,
            this.$terraformTemplateSourcesGeneratorTaskExecutor,
            this.$terraformSourcesImporterTaskExecutor,
            this.$terraformSourcesFormatterTaskExecutor,
            this.$terragruntConfigurationGeneratorTaskExecutor,
            this.$terraformTaskExecutor,
            this.$terraformRemoteStatesGeneratorTaskExecutor
        ].forEach((executor) => this.$taskExecutionService.registerExecutor(executor));
    }


}
