import cli from "chalk";
import * as util from "util";
import highlight from "cli-highlight";
import * as yaml from "yaml";
import CliLogLevels from "./CliLogLevels";
import {injectable} from "inversify";
import {ast} from "yaml";
const { Console } = require('console');
const fullConsole = new Console({
    stdout: process.stderr,
    stderr: process.stderr,
    colorMode: true
});
const quietConsole = new Console({
    stdout: process.stdout,
    stderr: process.stderr,
    colorMode: false
});
@injectable()
export default class Cli {

    private get console() {
        return this.quiet ? quietConsole : fullConsole;
    };

    constructor(
        public readonly levels = new CliLogLevels(),
        public quiet: boolean = false,
    ) {
    }

    log(...args: any[]) {
        this.console.log(...args);
    }

    logProcessOutput(...args: any[]) {
        this.console.log(...args.map((arg) => arg
            .trim()
            .split("\n")
            .map((line: string) => line.trim())
            .join("\n")
            .trim()
        ));
    }

    dump(object: any) {
        this.console.log(util.inspect(object, false, null, true /* enable colors */));
    }

    logInfo(...args: any[]) {
        if (this.quiet) return;
        this.console.log(cli.green(["  -", ...this.serializeArgs(args)].join(" ")));
    }

    logHeader(...args: any[]) {
        if (this.quiet) return;
        this.console.log(cli.green(["===", ...this.serializeArgs(args)].join(" ")));
    }

    logTaskStart(...args: any[]) {
        if (this.quiet) return;
        this.console.log(cli.white(["  $", ...this.serializeArgs(args)].join(" ")));
    }

    logLifecycleStart(...args: any[]) {
        if (this.quiet) return;
        this.console.log(cli.magentaBright(["——>", ...this.serializeArgs(args)].join(" ")));
    }

    logModuleProcessingStart(...args: any[]) {
        if (this.quiet) return;
        this.console.log(cli.magentaBright([">>>", ...this.serializeArgs(args)].join(" ")));
    }

    logTaskEnd(...args: any[]) {
        if (this.quiet) return;
        //this.console.log(cli.bold(["<", ...this.serializeArgs(args)].join(" ")));
    }

    logLifecycleEnd(...args: any[]) {
        if (this.quiet) return;
        //this.console.log(cli.bold(["<——", ...this.serializeArgs(args)].join(" ")));
    }

    logDebug(...args: any[]) {
        if (this.quiet || !this.levels.debug) return;
        const prefix = "   ";
        this.console.log(cli.dim([
                prefix,
                ...this.serializeArgs(args)
            ].join(" ").replace(/\n/g, "\n" + prefix)
        ));
    }

    logTrace(...args: any[]) {
        if (this.quiet || !this.levels.trace) return;
        const prefix = "   ";
        this.console.log(cli.dim([
                prefix,
                ...this.serializeArgs(args)
            ].join(" ").replace(/\n/g, "\n" + prefix)
        ));
    }

    logError(...args: any[]) {
        this.console.log("\n" + cli.red(["ERR", ...this.serializeArgs(args)].join(" ")));
    }

    logWarn(...args: any[]) {
        if (this.quiet) return;
        this.console.log(cli.yellow(["<!>", ...this.serializeArgs(args)].join(" ")));
    }

    logYamlObject(content: Object, force: boolean = false) {
        this.logYaml(yaml.stringify(content, {
            customTags: [
                {
                    class: Function,
                    tag: '!js',
                    stringify: (item: ast.Node) => {
                        return JSON.stringify((item as any).value.toString())
                    }
                } as any,
                {
                    class: Function,
                    tag: '!include.js',
                    stringify: (item: ast.Node) => {
                        return JSON.stringify((item as any).value.toString())
                    }
                } as any
            ]
        } as any), force);
    }

    logNamedYamlObject(content: Object, name: string = "", force: boolean = false) {
        this.logNamedYaml(yaml.stringify(content, {
            customTags: [
                {
                    class: Function,
                    tag: '!js',
                    stringify: (item: ast.Node) => {
                        return JSON.stringify((item as any).value.toString())
                    }
                } as any,
                {
                    class: Function,
                    tag: '!include.js',
                    stringify: (item: ast.Node) => {
                        return JSON.stringify((item as any).value.toString())
                    }
                } as any
            ]
        } as any), name, force);
    }

    logYaml(yaml: string, force: boolean = false) {
        if (!force && (this.quiet || !this.levels.trace)) {
            return;
        }
        const content = highlight(yaml, {language: 'yaml', ignoreIllegals: true});
        this.console.log(`---
${content.trim()}
`);
    }

    logNamedYaml(yaml: string, name: string = "", force: boolean = false) {
        if (!force && (this.quiet || !this.levels.trace)) {
            return;
        }
        const content = highlight(yaml, {language: 'yaml', ignoreIllegals: true});
        this.console.log(`---
# ${name}
${content.trim()}
`);
    }

    logHcl(hcl: string, name: string = "", force: boolean = false) {
        if (!force && (this.quiet || !this.levels.trace)) {
            return;
        }
        const content = highlight(hcl, {language: 'json', ignoreIllegals: true});
        this.console.log(`${name}
---
${content.trim()}
`);
    }

    private serializeArgs(args: any[]) {
        return args.map((arg) => {
            if (typeof arg == "string" || typeof arg == "number") {
                return arg as string;
            } else if (arg instanceof Array) {
                return "\n" + yaml.stringify(arg, {
                    customTags: [
                        {
                            class: Function,
                            tag: '!js',
                            stringify: (item: ast.Node) => {
                                return JSON.stringify((item as any).value.toString())
                            }
                        } as any,
                        {
                            class: Function,
                            tag: '!include.js',
                            stringify: (item: ast.Node) => {
                                return JSON.stringify((item as any).value.toString())
                            }
                        } as any
                    ]
                } as any);
            } else {
                return util.inspect(arg, false, null, true)
            }
        });
    }

};
