import Module from "../../domain/models/Module";
import Exception from "../error/Exception";
import * as Primitives from "../../domain/Primitives";
import LazyModule from "../../domain/models/LazyModule";

export default class ModulesRegistry {

    private readonly modules = new Map<string, Module>();
    private readonly lazyModules = new Map<string, LazyModule>();

    constructor(
        readonly context: Primitives.Serialized = {}
    ) {
    }

    get size(): number {
        return this.modules.size;
    }

    addLazyModules(modules: LazyModule[]) {
        modules.forEach((module) => this.setLazyModule(module.id, module));
    }

    removeLazyModule(moduleId: string) {
        this.lazyModules.delete(moduleId);
    }

    addModules(modules: Module[]) {
        modules.forEach((module) => this.setModule(module.id, module));
    }

    setModule(moduleId: string, module: Module) {
        this.removeLazyModule(moduleId);
        this.modules.set(moduleId, module);
    }

    setLazyModule(id: string, module: LazyModule) {
        this.lazyModules.set(id, module);
    }

    getModules(): Module[] {
        return Array.from(this.modules.values()).sort((a, b) => {
            if (a.id == ".") return -1;
            if (b.id == ".") return 1;
            return a.id.split("/") < b.id.split("/") ? -1 : 1
        });
    }

    getLazyModules(): LazyModule[] {
        return Array.from(this.lazyModules.values()).sort((a, b) => {
            if (a.id == ".") return -1;
            if (b.id == ".") return 1;
            return a.id.split("/") < b.id.split("/") ? -1 : 1
        });
    }

    getModuleIds() {
        return this.getModules().map((module) => module.id);
    }

    hasLazyModule(id: string) {
        return this.lazyModules.has(id);
    }

    hasModule(id: string) {
        return this.modules.has(id);
    }

    getModule(id: string): Module {
        if (!this.hasModule(id)) {
            throw new Exception(
                `Module '${id}' is not registered in registry!`,
                {ModuleId: id}
            );
        }
        return this.modules.get(id) as Module;
    }

    getLazyModule(id: string): LazyModule {
        if (!this.hasLazyModule(id)) {
            throw new Exception(
                `Module '${id}' is not registered in registry!`,
                {ModuleId: id}
            );
        }
        return this.lazyModules.get(id) as LazyModule;
    }

    forEach(callback: (module: Module, index: number, modules: Module[]) => void) {
        this.getModules().forEach(callback);
    }

}
