import {inject, injectable} from "inversify";
import ModulesRegistry from "../modules/ModulesRegistry";
import Cli from "../cli/Cli";
import * as Primitives from "../../domain/Primitives";
import ManifestsScanner from "../manifests/ManifestsScanner";
import Exception from "../error/Exception";
import ModelsBuilder from "../../domain/ModelsBuilder";
import * as Models from "../../domain/Models";
import ModulesSelector from "../../domain/ModulesSelector";
import * as Path from "path";
import * as Manifests from "../../domain/Manifests";
import * as _ from "lodash";
import * as fs from "fs";
import OpsApplicationConfiguration from "../../application/OpsApplicationConfiguration";
import * as inquirer from "inquirer";
import YamlService from "../yaml/YamlService";
import * as traverse from "traverse";
import YamlLazyValue from "../yaml/YamlLazyValue";
import * as semver from "semver";
import LazyModule from "../../domain/models/LazyModule";

const packageJson = require("./../../../package.json");

@injectable()
export default class ModulesService {

    private modulesRoot: string = "";
    private readonly registries: Map<string, ModulesRegistry>;
    private readonly cache: Map<string, Models.Module>;

    constructor(
        @inject(OpsApplicationConfiguration)
        private $applicationConfiguration: OpsApplicationConfiguration,
        @inject(Cli)
        private $cli: Cli,
        @inject(ManifestsScanner)
        private $manifestsScanner: ManifestsScanner,
        @inject(ModelsBuilder)
        private $modelsBuilder: ModelsBuilder,
        @inject(YamlService)
        private readonly $yaml: YamlService
    ) {
        this.registries = new Map();
        this.cache = new Map();
    }


    setModulesRoot(modulesRoot: string) {
        this.modulesRoot = modulesRoot;
    }

    getRegistry(id: string = "default"): ModulesRegistry {
        if (!this.hasRegistry(id)) {
            throw new Exception("Registry is not found", {RegistryId: id});
        }
        return this.registries.get(id) as ModulesRegistry;
    }

    hasRegistry(id: string) {
        return this.registries.has(id);
    }

    readModule(manifestPath: string, context: Primitives.Serialized = {}): Models.Module {
        if (this.cache.has(manifestPath)) {
            this.$cli.logTrace(`Reading module manifest '${manifestPath}' (cached)`);
            return this.cache.get(manifestPath) as Models.Module;
        }
        this.$cli.logTrace(`Reading module manifest '${manifestPath}'`);
        const modulePath = Path.dirname(Path.relative(this.modulesRoot, manifestPath));
        const moduleId = modulePath == "." ? "." : "./" + modulePath;
        const content = this.$yaml.parseFile(manifestPath);
        const manifest: Manifests.Manifest =
            _.merge(
                {},
                content,
                {
                    module: {
                        id: moduleId,
                        paths: {
                            module: moduleId,
                            manifest: manifestPath.toString(),
                            directory: Path.dirname(manifestPath.toString()),
                            unit: Path.basename(Path.dirname(manifestPath.toString())),
                            root: fs.realpathSync(this.modulesRoot)
                        }
                    }
                },
                context
            );
        try {
            const module = this.$modelsBuilder.buildModule(manifest);
            if (!semver.satisfies(packageJson.version, module.metadata.schema)) {
                throw new Exception(
                    "Module schema not supported by current ops version!",
                    {
                        ModuleId: moduleId,
                        File: manifestPath,
                        Schema: module.metadata.schema,
                        Version: packageJson.version,
                        Source: fs.readFileSync(manifestPath).toString()
                    }
                );
            }
            return module;
        } catch (error) {
            throw new Exception(
                "Unable to build module model from manifest!",
                {
                    ModuleId: moduleId,
                    File: manifestPath,
                    Source: fs.readFileSync(manifestPath).toString()
                },
                error
            );
        }
    }

    readInterpolatedModule(manifestPath: string, context: Primitives.Serialized = {}): Models.Module {
        if (this.cache.has(manifestPath)) {
            this.$cli.logTrace(`Reading interpolated module manifest '${manifestPath}' (cached)`);
            return this.cache.get(manifestPath) as Models.Module;
        }
        this.$cli.logTrace(`Reading interpolated module manifest '${manifestPath}'`);
        const modulePath = Path.dirname(Path.relative(this.modulesRoot, manifestPath));
        const moduleId = modulePath == "." ? "." : "./" + modulePath;
        const content = this.$yaml.parseJsonFile(manifestPath);
        const manifest: Manifests.Manifest =
            _.merge(
                {},
                content,
                {
                    module: {
                        id: moduleId,
                        paths: {
                            module: moduleId,
                            manifest: manifestPath.toString(),
                            directory: Path.dirname(manifestPath.toString()),
                            unit: Path.basename(Path.dirname(manifestPath.toString())),
                            root: fs.realpathSync(this.modulesRoot)
                        }
                    }
                },
                context
            );
        try {
            const module = this.$modelsBuilder.buildModule(manifest);
            if (!semver.satisfies(packageJson.version, module.metadata.schema)) {
                throw new Exception(
                    "Module schema not supported by current ops version!",
                    {
                        ModuleId: moduleId,
                        File: manifestPath,
                        Schema: module.metadata.schema,
                        Version: packageJson.version,
                        Source: fs.readFileSync(manifestPath).toString()
                    }
                );
            }
            this.cache.set(manifestPath, module);
            return module;
        } catch (error) {
            throw new Exception(
                "Unable to build module model from manifest!",
                {
                    ModuleId: moduleId,
                    File: manifestPath,
                    Source: fs.readFileSync(manifestPath).toString()
                },
                error
            );
        }
    }

    createRegistry(id: string, context: Primitives.Serialized = {}) {
        this.$cli.logTrace(`Initializing '${id}' modules registry:`);
        const registry = new ModulesRegistry(context);
        const manifestPaths = this.scanManifestPaths();
        const modules = manifestPaths.map((manifestPath) => {
            const modulePath = Path.dirname(Path.relative(this.modulesRoot, manifestPath));
            const moduleId = modulePath == "." ? "." : "./" + modulePath;
            return new LazyModule(
                moduleId,
                this.readModule.bind(this, manifestPath, context)
            )
        });
        registry.addLazyModules(modules);
        this.$cli.logTrace(`Loaded ${registry.size} modules!`);
        this.registries.set(id, registry);
        return registry;
    }

    queryModules(selector: ModulesSelector | string, registryId: string): string[] {
        if (typeof selector == "string") selector = {pattern: [selector]};
        const registry = this.getRegistry(registryId);
        const modules = registry.getModules();
        const lazyModules = registry.getLazyModules();
        this.$cli.logTrace("Querying modules:", selector, "Registry", registryId);
        return _.uniq(
            [...modules, ...lazyModules]
                .filter((target) => {
                    let filtered = false;
                    selector = selector as ModulesSelector;
                    if (selector.filter) {
                        filtered = true;
                        if (!selector.filter(target, selector)) return false;
                    }
                    if (selector.id) {
                        filtered = true;
                        const selectorIds = selector.id instanceof Array ?
                            selector.id :
                            [selector.id];
                        if (!selectorIds.includes(target.id)) return false;
                    }
                    if (selector.tag && target instanceof Models.Module) {
                        filtered = true;
                        const selectorTags = selector.tag instanceof Array ?
                            selector.tag :
                            [selector.tag];
                        const matches = selectorTags.filter((selectorTag) =>
                            target.metadata.tags.has(selectorTag));
                        // Expect all tags
                        if (matches.length != selectorTags.length) return false;
                    }
                    if (selector.ancestorOf) {
                        filtered = true;
                        const selectorAncestors = selector.ancestorOf instanceof Array ?
                            selector.ancestorOf :
                            [selector.ancestorOf];
                        const matches = selectorAncestors.filter((selectorAncestor) => {
                            let selectorAncestorIdChunks = selectorAncestor.split("/");
                            while (selectorAncestorIdChunks.length) {
                                selectorAncestorIdChunks = selectorAncestorIdChunks.slice(0, selectorAncestorIdChunks.length - 1);
                                if (target.id == selectorAncestorIdChunks.join("/")) return true;
                            }
                            return false;
                        });
                        if (!matches.length) return false;
                    }
                    if (selector.parentOf) {
                        filtered = true;
                        const selectorParents = selector.parentOf instanceof Array ?
                            selector.parentOf :
                            [selector.parentOf];
                        const matches = selectorParents.filter((selectorParent) => {
                            const selectorParentIdChunks = selectorParent.split("/");
                            const selectorParentId = selectorParentIdChunks.slice(0, selectorParentIdChunks.length - 1).join("/");
                            return target.id == selectorParentId;
                        });
                        if (!matches.length) return false;
                    }
                    if (selector.pattern) {
                        filtered = true;
                        const selectorPatterns = selector.pattern instanceof Array ?
                            selector.pattern :
                            [selector.pattern];
                        const matches = selectorPatterns.filter((selectorPattern) =>
                            target.id.match(new RegExp(
                                "^" + selectorPattern
                                    .replace(/\./g, "\.")
                                    .replace(/\//g, "\/")
                                    .replace(/[*]{2}/g, ".+")
                                    .replace(/[*]/g, "[a-zA-Z0-9_-]+") + "$",
                                "g"
                            ))
                        );
                        if (!matches.length) return false;
                    }
                    if (!filtered) {
                        throw new Exception("Invalid selector", selector);
                    }
                    return true;
                })
                .map((module) => module.id)
                .sort((a, b) => {
                    if (a == ".") return -1;
                    if (b == ".") return 1;
                    return a.split("/") < b.split("/") ? -1 : 1
                })
        );
    }

    hasModule(moduleId: string, registryId: string) {
        const registry = this.getRegistry(registryId);
        return registry.hasModule(moduleId) || registry.hasLazyModule(moduleId);
    }

    queryModule(selector: ModulesSelector | string, registry: string): string {
        return this.queryModules(selector, registry)[0];
    }

    async getModule(moduleId: string, registryId: string) {
        this.$cli.logTrace(`Getting module '${moduleId}' (${registryId})...`)
        const module = this.lookupModule(moduleId, registryId);
        if (module.metadata.interpolation == "completed") {
            return module;
        }
        if (module.metadata.interpolation != "none") {
            throw new Error(`Module '${module.id}' interpolation in progress!`);
        }
        this.$cli.logTrace(`Loading module '${moduleId}' (${registryId})...`);
        module.metadata.interpolation = "in-progress";
        if (module.id != ".") {
            let ancestorId = this
                .queryModules({ancestorOf: module.id}, registryId)
                .reverse()[0];
            if (this.hasModule(ancestorId, registryId)) {
                let ancestor = this.lookupModule(ancestorId, registryId);
                switch (ancestor.metadata.interpolation) {
                    case "completed":
                        if (!module.metadata.parents.has(ancestor.id)) {
                            this.$cli.logTrace(`Inheriting ancestor '${ancestor.id}' –> '${module.id}'...`);
                            module.inheritFrom(ancestor);
                        }
                        break;
                    case "none":
                        ancestor = await this.getModule(ancestorId, registryId);
                        if (!module.metadata.parents.has(ancestor.id)) {
                            this.$cli.logTrace(`Inheriting ancestor '${ancestor.id}' –> '${module.id}'...`);
                            module.inheritFrom(ancestor);
                        }
                        break;
                    default:
                        this.$cli.logTrace(`Inheriting ancestor '${ancestor.id}' –> '${module.id}' is pending...`);
                        break;
                }
            }
        }
        //
        const providers = this.queryModules({
            filter: (target) => {
                if (target instanceof Models.Module) {
                    if (!target.metadata.inherit.into.selector.tag) {
                        return false;
                    }
                    const selectorTags = target.metadata.inherit.into.selector.tag instanceof Array ?
                        target.metadata.inherit.into.selector.tag :
                        [target.metadata.inherit.into.selector.tag];
                    const matches = selectorTags.filter((selectorTag) =>
                        module.metadata.tags.has(selectorTag));
                    return matches.length == selectorTags.length;
                } else {
                    return false;
                }
            }
        }, registryId);
        //
        if (providers.length) {
            for (let providerIndex = 0; providerIndex < providers.length; providerIndex++) {
                const providerId = providers[providerIndex];
                let provider = this.lookupModule(providerId, registryId);
                switch (provider.metadata.interpolation) {
                    case "completed":
                        if (!module.metadata.parents.has(provider.id)) {
                            this.$cli.logTrace(`Inheriting provider '${provider.id}' –> '${module.id}'...`);
                            module.inheritFrom(provider);
                        }
                        break;
                    case "none":
                        provider = await this.getModule(providerId, registryId);
                        if (!module.metadata.parents.has(provider.id)) {
                            this.$cli.logTrace(`Inheriting provider '${provider.id}' –> '${module.id}'...`);
                            module.inheritFrom(provider);
                        }
                        break;
                    default:
                        break;
                }
            }
        }
        //
        if (!module.metadata.virtual) {
            const moduleInputs = module.inputs.entries;
            for (let moduleInputIndex = 0; moduleInputIndex < moduleInputs.length; moduleInputIndex++) {
                const moduleInput = moduleInputs[moduleInputIndex][1];
                this.$cli.logDebug((moduleInput.required ? "Required" : "Optional") + " input: ", moduleInput.name);
                if (this.$applicationConfiguration.context.inputs.hasOwnProperty(moduleInput.name)) {
                    moduleInput.value = this.$applicationConfiguration.context.inputs[moduleInput.name];
                    moduleInput.provided = true;
                    this.$cli.logTrace("Value provided from ops command line arguments:", moduleInput.value);
                }
                if (!moduleInput.provided && moduleInput.prompt) {
                    moduleInput.value = ((await inquirer.prompt([
                        {
                            name: "input",
                            message: moduleInput.prompt.message,
                            type: moduleInput.prompt.type,
                            choices: (typeof moduleInput.prompt.choices != 'function' ?
                                moduleInput.prompt.choices :
                                (function () {
                                    return (moduleInput.prompt.choices as Function)(module.serialize());
                                }())),
                            default: (typeof moduleInput.prompt.defaults != 'function' ?
                                moduleInput.prompt.defaults :
                                (function () {
                                    return (moduleInput.prompt.defaults as Function)(module.serialize());
                                }()))
                        }
                    ])) as any).input;
                    this.$applicationConfiguration.context.inputs[moduleInput.name] = moduleInput.value;
                    moduleInput.provided = true;
                }
                if (!moduleInput.provided && moduleInput.required) {
                    this.$cli.logError("Required input parameter", moduleInput.name, "is not provided!");
                    process.exit(1);
                }
            }
        }
        const cwd = module.paths.directory;
        const registry = this.getRegistry(registryId);
        const manifestObject = _.merge(
            {
                module: module.serialize(),
                application: {
                    env: process.env,
                    context: this.$applicationConfiguration.context,
                    version: this.$applicationConfiguration.version,
                    requiredCommands: this.$applicationConfiguration.requiredCommands
                }
            },
            registry.context
        );
        if (module.metadata.virtual) {
            traverse(manifestObject.module).forEach(function (element) {
                if (element instanceof YamlLazyValue) {
                    this.update("!lazy " + element.value);
                }
            });
        } else {
            traverse(manifestObject.module).forEach(function (element) {
                if (element instanceof YamlLazyValue) {
                    this.update(element.value);
                }
            });
        }
        let manifestYaml = this.$yaml.stringify(manifestObject);
        if (!module.metadata.virtual) {
            manifestYaml = manifestYaml
                .replace(/"<<-": '!lazy /g, '<<: \'')
                .replace(/<<-: '!lazy /g, '<<: \'')
                .replace(/"<<-": "!lazy /g, '<<: \"')
                .replace(/<<-: "!lazy /g, '<<: \"')
                .replace(/: '!lazy /g, ': \'')
                .replace(/: "!lazy /g, ': \"')
                .replace(/- '!lazy /g, '- \'')
                .replace(/- "!lazy /g, '- \"')
                .replace(/"<<-": !lazy /g, '<<: ')
                .replace(/<<-: !lazy /g, '<<: ')
                .replace(/: !lazy /g, ': ')
                .replace(/- !lazy /g, '- ')
                .replace(/"<<-": /g, "<<: ")
                .replace(/<<-: /g, "<<: ")
                .replace(/\s"!lazy /g, ' "');
        }
        if (providers.length) {
            manifestYaml = manifestYaml
                .replace(/<<-: '!direct-inheritor /g, '<<: \'')
                .replace(/<<-: "!direct-inheritor /g, '<<: \"')
                .replace(/: '!direct-inheritor /g, ': \'')
                .replace(/: "!direct-inheritor /g, ': \"')
                .replace(/- '!direct-inheritor /g, '- \'')
                .replace(/- "!direct-inheritor /g, '- \"')
                .replace(/\s"!direct-inheritor /g, ' "');
        }
        const rawFile = `${cwd}/~~module.manifest.raw.yaml`;
        const interpolatedFile = `${cwd}/~~module.manifest.interpolated.yaml`;
        try {
            fs.writeFileSync(rawFile, manifestYaml);
            const interpolatedYaml = this.$yaml.spiffFiles([rawFile], {
                asJson: true,
                env: module.env.filter((value) => !(value instanceof YamlLazyValue)).serialize()
            });
            fs.writeFileSync(interpolatedFile, interpolatedYaml);
            const interpolated = this.readInterpolatedModule(interpolatedFile, {
                module: {
                    metadata: {
                        $interpolation: "completed",
                        $parents: module.metadata.parents
                    }
                }
            });
            registry.setModule(module.id, interpolated);
            this.$cli.logDebug(`Loaded module '${module.id}'...`);
            return interpolated;
        } catch (error) {
            throw new Exception(`Unable to interpolate module!`, {
                ModuleId: module.id
            }, error);
        }
    }

    private scanManifestPaths() {
        this.$cli.logTrace(`Scanning module in '${this.modulesRoot}'...`);
        return this.$manifestsScanner.find(this.modulesRoot);
    }

    private lookupModule(moduleId: string, registryId: string): Models.Module {
        const registry = this.getRegistry(registryId);
        if (registry.hasLazyModule(moduleId)) {
            this.$cli.logTrace(`Loading lazy module '${moduleId}' (${registryId})...`)
            const lazyModule = registry.getLazyModule(moduleId);
            const module = lazyModule.read();
            registry.setModule(moduleId, module);
            if (module.metadata.autoload.size) {
                const autoloadModuleIds = this.queryModules({
                    pattern: Array.from(module.metadata.autoload)
                }, registryId);
                autoloadModuleIds.forEach((moduleId) => {
                    this.$cli.logTrace(`Auto-loading module '${moduleId}' (${registryId})...`)
                    this.lookupModule(moduleId, registryId);
                })
            }
            return module;
        } else if (registry.hasModule(moduleId)) {
            return registry.getModule(moduleId);
        } else {
            throw new Exception("Module is not found", {
                ModuleId: moduleId,
                RegistryId: registryId
            });
        }
    }

}
