import TaskExecutor from "./TaskExecutor";
import {injectable} from "inversify";
import Exception from "../error/Exception";
import * as Models from "../../domain/Models";
import * as tmp from "tmp";
import * as fs from "fs";
import OpsApplicationContext from "../../application/OpsApplicationContext";

@injectable()
export default class TaskExecutionService {

    protected readonly $executors: Map<string, TaskExecutor> = new Map();
    async executeTask(
        task: Models.ModuleTask,
        context: OpsApplicationContext,
        targetModule: Models.Module,
        bindingModule?: Models.Module,
    ): Promise<any> {
        if (!task.metadata.executor) {
            throw new Exception(`Task has no executor defined!`, {
                TaskName: task.name,
                OwnerModuleId: targetModule.id
            });
        }
        const taskExecutor = this.getExecutor(task.metadata.executor);
        const opsContextFile = tmp.fileSync();
        fs.writeFileSync(opsContextFile.name, JSON.stringify({
            module: targetModule.serialize(),
            binding: bindingModule ? bindingModule.serializeAsBinding() : {}
        }));
        try {
            await taskExecutor.execute(task, {
                env: {
                    OPS_CONTEXT_FILE: opsContextFile.name,
                    OPS_TARGET_MODULE: module.id,
                    OPS_BINDING_MODULE: bindingModule ? bindingModule.id : undefined,
                    OPS_PATH_MODULES: context.modulesRoot
                }
            }, targetModule, bindingModule);
        } finally {
            opsContextFile.removeCallback()
        }

    }

    registerExecutor(instance: TaskExecutor) {
        this.$executors.set(instance.name, instance);
    }

    hasExecutor(name: string) {
        return this.$executors.has(name);
    }

    getExecutor(name: string): TaskExecutor {
        if (!this.hasExecutor(name))
            throw new Exception(`Task executor is not found!`, {
                TaskName: name
            });
        return this.$executors.get(name) as TaskExecutor;
    }

}
