import TaskExecutor from "../../TaskExecutor";
import * as glob from "glob";
import * as fs from "fs";
import {inject, injectable} from "inversify";
import Cli from "../../../cli/Cli";
import * as Models from "../../../../domain/Models";
import FSTaskConfigurationActions from "./FSTaskConfigurationActions";
import {execFileSync} from "child_process";
import YamlService from "../../../yaml/YamlService";
import TaskExecutorContext from "../../TaskExecutorContext";

@injectable()
export default class FSTaskExecutor
    implements TaskExecutor {

    public readonly name: string = "fs";

    constructor(
        @inject(Cli)
        protected $cli: Cli,
        @inject(YamlService)
        protected readonly $yaml: YamlService
    ) {

    }

    async execute(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        if (!task.configuration) {
            throw new Error;
        }
        this.$cli.logTrace("Executing task with configuration:", task.configuration);
        switch (task.configuration["action"]) {
            case FSTaskConfigurationActions.REMOVE:
                this.removeTargets(this.findTargets(task.configuration["selectors"], {
                    sync: true,
                    absolute: true,
                    realpath: true,
                    cwd: module.paths.directory
                }));
                break;
            case FSTaskConfigurationActions.WRITE:
                if (!task.configuration["content"] && task.configuration["content-from-exec"]) {
                    const execConfig = task.configuration["content-from-exec"];
                    task.configuration["content"] = execFileSync(
                        execConfig.command,
                        execConfig.args,
                        {
                            cwd: execConfig.cwd || process.cwd(),
                            timeout: execConfig.timeout || undefined,
                            env: execConfig.env || process.env
                        }
                    ).toString("utf-8");
                }
                this.writeContentToTargets(
                    this.normalizeTargets(task.configuration["selectors"], module.paths.directory),
                    task.configuration["content"],
                    task.configuration["content-type"]
                );
                break;
            default:
                throw new Error;
        }
    }

    private normalizeTargets(selectors: string | string[], cwd: string): string[] {
        if (typeof selectors == "string") {
            selectors = [selectors]
        }
        return selectors;
    }

    private findTargets(selectors: string | string[], options: glob.IOptions): string[] {
        if (typeof selectors == "string") {
            selectors = [selectors]
        }
        const targets = new Set<string>();
        selectors.forEach((selector) =>
            glob.sync(selector, options).forEach((target) => targets.add(target))
        );
        return Array.from(targets);
    }

    private removeTargets(targets: string[]) {
        if (!targets.length) {
            this.$cli.logDebug("No files to remove!");
        } else {
            this.$cli.logTrace("Removing files:", targets);
            let removalFiles = 0;
            let removalDirectories = 0;
            targets.forEach((target) => {
                if (fs.lstatSync(target).isDirectory()) {
                    [
                        glob.sync("*", {
                            sync: true,
                            absolute: true,
                            realpath: true,
                            cwd: target
                        }),
                        glob.sync(".*", {
                            sync: true,
                            absolute: true,
                            realpath: true,
                            cwd: target
                        })
                    ].forEach((files) => {
                        if (files.length) {
                            const quiet = this.$cli.quiet;
                            this.$cli.quiet = true;
                            this.removeTargets(files);
                            this.$cli.quiet = quiet;
                        }
                    });
                    fs.rmdirSync(target);
                    removalDirectories++;
                } else {
                    fs.unlinkSync(target);
                    removalFiles++;
                }
            });
            this.$cli.logInfo(`Removed ${removalFiles} files!`);
            if (removalDirectories) {
                this.$cli.logInfo(`Removed ${removalDirectories} directories!`);
            }
        }
    }

    private writeContentToTargets(targets: string[], content: any, contentType: string) {
        if (typeof content != "string") {
            switch (contentType) {
                case "yaml":
                    content = this.$yaml.stringify(content);
                    break;
                default:
                    throw new Error;
            }
        }
        content = content.toString();
        targets.forEach((target) => {
            this.$cli.logDebug(`Writing (${contentType}) content (${content.length} bytes) into: ${target}`);
            fs.writeFileSync(target, content);
        });
    }
}
