import * as Primitives from "../../../../domain/Primitives";
import Exception from "../../../error/Exception";
import * as Models from "../../../../domain/Models";
import * as _ from "lodash";

const Validator = require("fastest-validator");

export default class TerraformRemoteStatesGeneratorTaskConfiguration {
    constructor(
        readonly outputFileName: string,
        readonly remotes: {
            name: string,
            state: string,
            directory: string,
            defaults: Primitives.Serialized
        }[],
        readonly stateRegion: string,
        readonly stateDirectory: string,
        readonly stateBucketName: string,
        readonly stateBucketEncrypt: boolean,
        readonly stateDynamoDBTable: string,
        readonly stateCredentialsAccessKey: string,
        readonly stateCredentialsSecretKey: string
    ) {
    }

    static parse(
        configuration: Primitives.Serialized,
        module: Models.Module,
        binding: Models.Module
    ): TerraformRemoteStatesGeneratorTaskConfiguration {
        if (!configuration) throw new Error("Missing configuration!");
        configuration = _.merge({
            output: {
                file: {
                    name: "~states.tf"
                }
            },
            state: {
                directory: module.context.environment.terraform.state.directory,
                region: module.context.environment.terraform.backend.s3.region,
                dynamodb: {
                    table: module.context.environment.terraform.backend.s3.dynamodb.table
                },
                bucket: {
                    name: module.context.environment.terraform.backend.s3.bucket.name,
                    encrypt: module.context.environment.terraform.backend.s3.bucket.encrypt
                },
                credentials: {
                    "access-key": module.context.environment.terraform.backend.s3.credentials["access-key"],
                    "secret-key": module.context.environment.terraform.backend.s3.credentials["secret-key"]
                }
            }
        }, configuration);
        const validator: any = new Validator();
        const errors = validator.validate(configuration, {
            output: {
                type: "object",
                props: {
                    file: {
                        type: "object",
                        props: {
                            name: {type: "string"}
                        }
                    }
                }
            },
            state: {
                type: "object",
                props: {
                    directory: {type: "string"},
                    region: {type: "string"},
                    bucket: {
                        type: "object",
                        props: {
                            name: {type: "string"},
                            encrypt: {type: "boolean"}
                        }
                    },
                    credentials: {
                        type: "object",
                        props: {
                            "access-key": {type: "string"},
                            "secret-key": {type: "string"}
                        }
                    }
                }
            }
        } as any);
        if (configuration.remotes instanceof Array) {
            configuration.remotes = configuration.remotes.map((remote) => {
                return {
                    name: remote.replace(/\./g, "-"),
                    state: remote,
                    directory: configuration.state.directory,
                    defaults: {}
                };
            });
        } else if (configuration.remotes instanceof Object) {
            configuration.remotes = Object.keys(configuration.remotes).map((remote) => {
                const config = _.merge({
                    name: remote.replace(/\./g, "-"),
                    directory: configuration.state.directory,
                    defaults: {}
                }, configuration.remotes[remote]);
                return config;
            });
        } else {
            throw new Exception(".configuration.remotes is invalid!");
        }
        if (errors !== true)
            throw new Exception("Invalid task configuration:", errors);
        return new TerraformRemoteStatesGeneratorTaskConfiguration(
            configuration.output.file.name,
            configuration.remotes,
            configuration.state.region,
            configuration.state.directory,
            configuration.state.bucket.name,
            configuration.state.bucket.encrypt,
            configuration.state.dynamodb.table,
            configuration.state.credentials["access-key"],
            configuration.state.credentials["secret-key"]
        );
    }
}
