import TaskExecutor from "../../TaskExecutor";
import * as fs from "fs";
import {inject, injectable} from "inversify";
import Cli from "../../../cli/Cli";
import * as Models from "../../../../domain/Models";
import TerraformRemoteStatesGeneratorTaskConfiguration from "./TerraformRemoteStatesGeneratorTaskConfiguration";
import TaskExecutorContext from "../../TaskExecutorContext";

@injectable()
export default class TerraformRemoteStatesGeneratorTaskExecutor
    implements TaskExecutor {

    public readonly name: string = "terraform-remote-states-generator";

    constructor(
        @inject(Cli) protected $cli: Cli
    ) {
    }

    serializeValue(value: any): string {
        if (value instanceof Array) {
            return JSON.stringify(value.map((value) => {
                return this.serializeValue(value)
            }))
        } else if (value instanceof Object) {
            return "{" + Object.keys(value).map((key) => {
                return `${key} = ${this.serializeValue(value[key])}`;
            }).join("\n") + "}"
        } else {
            return JSON.stringify(value);
        }
    }

    async execute(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        const configuration = TerraformRemoteStatesGeneratorTaskConfiguration.parse(
            task.configuration,
            module,
            binding
        );
        this.$cli.logNamedYamlObject(configuration, `Task '${task.name}' configuration`);
        if (!configuration.remotes.length) {
            this.$cli.logDebug("No remotes specified!");
            return;
        }
        const cwd = module.paths.directory;
        const source = configuration.remotes.map((remote) => {
            return `
data "terraform_remote_state" "${remote.name}" {
  backend = "s3"
  defaults = ${this.serializeValue(remote.defaults)}
  config {
    key            = "${remote.directory}/${remote.state}.tfstate"
    bucket         = "${configuration.stateBucketName}"
    region         = "${configuration.stateRegion}"
    encrypt        = "${configuration.stateBucketEncrypt ? "true" : "false"}"
    dynamodb_table = "${configuration.stateDynamoDBTable}"
    access_key     = "${configuration.stateCredentialsAccessKey}"
    secret_key     = "${configuration.stateCredentialsSecretKey}"
  }
}
`
        }).join("\n\n");
        fs.writeFileSync(`${cwd}/${configuration.outputFileName}`, source);
    }

}
