import TaskExecutor from "../../TaskExecutor";
import * as fs from "fs";
import * as yaml from "yaml";
import {execFileSync} from "child_process";
import * as _ from "lodash";
import TaskExecutionError from "../../TaskExecutionError";
import {inject, injectable} from "inversify";
import Cli from "../../../cli/Cli";
import * as Models from "../../../../domain/Models";
import * as Primitives from "../../../../domain/Primitives";
import TerraformSectionSourcesGeneratorTaskConfiguration from "./TerraformSectionSourcesGeneratorTaskConfiguration";
import YamlService from "../../../yaml/YamlService";
import TaskExecutorContext from "../../TaskExecutorContext";
import YamlLazyValue from "../../../yaml/YamlLazyValue";

@injectable()
export default class TerraformSectionSourcesGeneratorTaskExecutor
    implements TaskExecutor {

    public readonly name: string = "terraform-section-sources-generator";

    constructor(
        @inject(Cli)
        protected $cli: Cli,
        @inject(YamlService)
        private readonly $yaml: YamlService
    ) {
    }

    async execute(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        if (!task.configuration) throw new Error;
        const configuration = TerraformSectionSourcesGeneratorTaskConfiguration.parse(task.configuration);
        this.$cli.logNamedYamlObject(configuration, `Task '${task.name}' configuration`);
        if (!module.templates.has(configuration.templateSection)) {
            if (task.metadata.enabled) {
                throw new TaskExecutionError(
                    `No template '${configuration.templateSection}' defined! ` +
                    `Found: ${JSON.stringify(module.templates.getNames())}!`);
            } else {
                this.$cli.logDebug(`No template '${configuration.templateSection}' defined! Skipping execution!`);
                return;
            }
        }
        const cwd = module.paths.directory;
        const inputFile = `${cwd}/~~terraform.sources.input.yaml`;
        let templateSection = module.templates.get(configuration.templateSection);
        if (typeof templateSection != "string") {
            templateSection = this.$yaml.stringify(templateSection);
        }
        fs.writeFileSync(inputFile, templateSection);
        const variablesData: Primitives.Serialized = {
            module: module.serialize()
        };
        const variablesStructure: Primitives.Serialized = {
            module: "(( &temporary ))"
        };
        if (binding) {
            variablesData.binding = binding.serialize();
            variablesStructure.binding = "(( &temporary ))";
        }
        const variablesYaml = this.$yaml.stringify(variablesData);
        const variablesFile = `${cwd}/~~terraform.sources.variables.yaml`;
        const templateData = _.merge(yaml.parse(fs.readFileSync(inputFile).toString()), variablesStructure);
        const templateYaml = this.$yaml.stringify(templateData);
        const templateFile = `${cwd}/~~terraform.sources.template.yaml`;
        this.$cli.logDebug(`Interpolating ${module.id} ${configuration.templateSection}:`);
        fs.writeFileSync(templateFile, templateYaml);
        fs.writeFileSync(variablesFile, variablesYaml);
        const resultStructure = this.$yaml.parseSpiffFiles([templateFile, variablesFile], {
            env: module.env.filter((value) => !(value instanceof YamlLazyValue)).serialize(),
            cwd: cwd
        });
        const resultJSON = JSON.stringify(resultStructure, null, 4);
        const resultFile = `${cwd}/~~terraform.source.result.yaml`;
        this.$cli.logDebug(`Generating ${module.id} ${configuration.templateSection}:`);
        fs.writeFileSync(resultFile, resultJSON);
        const outputHCL = execFileSync(
            "bash",
            ['-c', `json2hcl < ${resultFile}`],
            {
                cwd: cwd,
                env: _.merge({}, process.env, module.env.toHash())
            }
        )
            .toString()
            .split("\n")
            .map((line) => line.replace(/^(\s*)?"([a-z_-]*)"/g, "$1$2"))
            .join("\n")
            .replace(/\n\n/g, "\n");
        fs.writeFileSync(`${cwd}/${configuration.outputFileName}`, outputHCL);
        this.$cli.logDebug(`Generated terraform sources '${configuration.outputFileName}'`);
    }

}
