import TaskExecutor from "../../TaskExecutor";
import * as fs from "fs";
import * as _ from "lodash";
import * as glob from "glob";
import TerraformSourcesImporterTaskConfiguration from "./TerraformSourcesImporterTaskConfiguration";
import {inject, injectable} from "inversify";
import Cli from "../../../cli/Cli";
import * as Models from "../../../../domain/Models";
import TaskExecutionException from "../../TaskExecutionError";
import ModulesService from "../../../modules/ModulesService";
import TaskExecutorContext from "../../TaskExecutorContext";

@injectable()
export default class TerraformSourcesImporterTaskExecutor
    implements TaskExecutor {

    public readonly name: string = "terraform-sources-importer";

    constructor(
        @inject(Cli) protected $cli: Cli,
        @inject(ModulesService) protected $modulesService: ModulesService,
    ) {
    }

    async execute(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        const configuration = TerraformSourcesImporterTaskConfiguration.parse(task.configuration);
        this.$cli.logNamedYamlObject(configuration, `Task '${task.name}' configuration`);
        const cwd = module.paths.directory;
        const importedTerraformSources: string[] = [];
        const importedTerraformVariables = {};
        this.$modulesService.createRegistry("task", {
            binding: binding.serializeAsBinding()
        });
        const modules = this.$modulesService.queryModules(configuration.selector, "task");
        this.$cli.logDebug(
            "Importing terraform sources from modules: ",
            modules
        );
        for (let moduleIndex = 0; moduleIndex < modules.length; moduleIndex++) {
            const moduleId = modules[moduleIndex];
            const module = await this.$modulesService.getModule(moduleId, "task");
            const sourceFiles = glob.sync("*.tf", {
                cwd: module.paths.directory,
                absolute: true
            });
            sourceFiles.forEach((file) => {
                this.$cli.logDebug(`- ${file}`);
                importedTerraformSources.push(
                    `/* ${file} */`,
                    fs.readFileSync(file).toString()
                );
            });
            const jsonVariableFiles = glob.sync("*.tfvars.json", {
                cwd: module.paths.directory,
                absolute: true
            });
            jsonVariableFiles.forEach((file) => {
                this.$cli.logDebug(`- ${file}`);
                _.merge(importedTerraformVariables, JSON.parse(fs.readFileSync(file).toString()));
            });
        }
        fs.writeFileSync(`${cwd}/${configuration.outputFilePrefix}.tf`, importedTerraformSources.join("\n\n"));
        if (Object.keys(importedTerraformVariables).length) {
            fs.writeFileSync(
                `${cwd}/${configuration.outputFilePrefix}.tfvars.json`,
                JSON.stringify(importedTerraformVariables, null, 4)
            );
        }
        if (modules.length) {
            this.$cli.logInfo(`Imported terraform sources from ${modules.length} modules!`);
        } else {
            throw new TaskExecutionException(`No modules for terraform sources import!`, {
                Configuration: task.configuration
            });
        }
    }

}
