import TaskExecutor from "../../TaskExecutor";
import TerraformTaskConfiguration from "./TerraformTaskConfiguration";
import {execFile, execFileSync} from "child_process";
import * as _ from "lodash";
import * as ora from "ora";
import TerraformTaskConfigurationActions from "./TerraformTaskConfigurationActions";
import TaskExecutionException from "../../TaskExecutionError";
import TerraformTaskMoveStateResourcesConfiguration from "./TerraformTaskMoveStateResourcesConfiguration";
import {inject, injectable} from "inversify";
import Cli from "../../../cli/Cli";
import TerraformTaskMoveStateFileConfiguration from "./TerraformTaskMoveStateFileConfiguration";
import * as Primitives from "../../../../domain/Primitives";
import * as Models from "../../../../domain/Models";
import TerraformTaskPruneStateResourcesConfiguration from "./TerraformTaskPruneStateResourcesConfiguration";
import Exception from "../../../error/Exception";
import * as fs from "fs";
import TaskExecutorContext from "../../TaskExecutorContext";
import OpsApplicationConfiguration from "../../../../application/OpsApplicationConfiguration";

const stripAnsi = require("strip-ansi");

type Pattern = {
    selector?: string | RegExp | (string | RegExp)[],
    action: ((
        data: string,
        event: {
            channelName: string,
            spinner: {
                stop: (() => void),
                clear: (() => void)
            }
        }
    ) => Promise<boolean>);
}

@injectable()
export default class TerraformTaskExecutor
    implements TaskExecutor {

    public readonly name: string = "terraform";

    constructor(
        @inject(Cli) protected $cli: Cli,
        @inject(OpsApplicationConfiguration)
        readonly $configuration: OpsApplicationConfiguration,
    ) {

    }

    async execute(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        if (!task.configuration) throw new Error;
        const configuration = new TerraformTaskConfiguration(
            task.configuration["action"],
            task.configuration["retries"]
        );
        if (!configuration.action) {
            throw new TaskExecutionException(`Missing action in configuration!`);
        }
        const retry = async (callback: () => Promise<number | boolean>) => {
            let retry = 1;
            while (retry <= configuration.retries) {
                try {
                    const status = await callback();
                    if (-1 == status) {
                        retry++;
                        if (retry > configuration.retries) {
                            throw new Exception("Max retries exceeded!");
                        }
                        this.$cli.logInfo(`Retrying ${retry}/${configuration.retries}...`);
                        continue;
                    }
                    break;
                } catch (error) {
                    [
                        "unknown variable",
                        "invalid or unknown key",
                        "unknown resource",
                        "referenced in variable",
                        "define it with a 'variable' block",
                        "ValidationError",
                        "does not have attribute",
                        "Creation of service was not idempotent",
                        "parse error at",
                        "should be type map, got type string in",
                        "must be a single value, not a map"
                    ].forEach((pattern) => {
                        if (stripAnsi(error.message).includes(pattern)) {
                            throw new Exception("Terraform fatal failure!", {}, error);
                        }
                    });
                    retry++;
                    if (retry > configuration.retries) throw error;
                    this.$cli.logInfo(`Retrying ${retry}/${configuration.retries}...`);
                }
            }
        };
        switch (configuration.action) {
            case TerraformTaskConfigurationActions.PLAN:
                await retry(async () => {
                    return this.plan(task.configuration, context, module, binding);
                });
                break;
            case TerraformTaskConfigurationActions.OUTPUT:
                await retry(async () => {
                    return this.output(task.configuration, context, module, binding);
                });
                break;
            case TerraformTaskConfigurationActions.VALIDATE:
                await this.validate(task.configuration, context, module, binding);
                break;
            case TerraformTaskConfigurationActions.DESTROY:
                await retry(async () => {
                    return this.destroy(task.configuration, context, module, binding);
                });
                break;
            case TerraformTaskConfigurationActions.DEPLOY:
                await retry(async () => {
                    return this.deploy(task.configuration, context, module, binding);
                });
                break;
            case TerraformTaskConfigurationActions.STATE_RESOURCES_PRUNE:
                await retry(async () => {
                    return this.pruneStateResources(task, context, module, binding);
                });
                break;
            case TerraformTaskConfigurationActions.STATE_RESOURCES_MOVE:
                await retry(async () => {
                    return this.moveStateResources(task, context, module, binding);
                });
                break;
            case TerraformTaskConfigurationActions.STATE_FILE_MOVE:
                await retry(async () => {
                    return this.moveStateFile(task, context, module, binding);
                });
                break;
            default:
                throw new TaskExecutionException(`Unknown action '${configuration.action}'!`);
        }
    }

    private async hasBucketFile(configuration: {
        path: string,
        bucket: {
            region: string,
            name: string,
            credentials: {
                accessKey: string,
                secretKey: string
            }
        }
    }, module: Models.Module) {
        try {
            execFileSync(
                this.$configuration.context.cmdAWS,
                [
                    "--output",
                    "json",
                    "s3",
                    "ls",
                    "s3://" + configuration.bucket.name + "/" + configuration.path
                ],
                {
                    timeout: 30 * 60 * 1000,
                    cwd: module.paths.directory,
                    env: _.merge({}, process.env, module.env.toHash(), {
                        AWS_ACCESS_KEY_ID: configuration.bucket.credentials.accessKey,
                        AWS_SECRET_ACCESS_KEY: configuration.bucket.credentials.secretKey,
                        AWS_DEFAULT_REGION: configuration.bucket.region
                    })
                }
            );
            return true;
        } catch (error) {
            if (error.stderr.toString() == "" && error.stdout.toString() == "") {
                return false;
            }
            throw new TaskExecutionException(
                "Unable to list state file!",
                configuration,
                error
            );
        }
    }

    private async moveStateFile(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        const configuration = TerraformTaskMoveStateFileConfiguration.parse(
            task.configuration,
            module,
            binding
        );
        this.$cli.logNamedYamlObject(configuration, `Task '${task.name}' configuration`);
        if (!configuration.stateFileSource) {
            this.$cli.logDebug("No source state file specified!");
            return false;
        }
        this.$cli.logInfo(`Validating terraform state file location: ${configuration.stateFileSource}`);
        const newStateExists = await this.hasBucketFile({
            path: configuration.stateFileTarget,
            bucket: {
                name: configuration.stateBucketName,
                region: configuration.stateRegion,
                credentials: {
                    accessKey: configuration.stateCredentialsAccessKey,
                    secretKey: configuration.stateCredentialsSecretKey
                }
            }
        }, module);
        if (newStateExists) {
            this.$cli.logWarn("State file already migrated!");
            this.$cli.logInfo("Target path:", configuration.stateFileTarget);
            return false;
        }
        const oldStateExists = await this.hasBucketFile({
            path: configuration.stateFileSource,
            bucket: {
                name: configuration.stateBucketName,
                region: configuration.stateRegion,
                credentials: {
                    accessKey: configuration.stateCredentialsAccessKey,
                    secretKey: configuration.stateCredentialsSecretKey
                }
            }
        }, module);
        if (!oldStateExists) {
            this.$cli.logWarn("Source state file is not found! Fresh deployment?", {
                path: configuration.stateFileSource
            });
            return false;
        }
        this.$cli.logWarn(`Migrating (by ${configuration.stateFileMethod}) state file...`);
        this.$cli.logDebug("Path:", configuration.stateFileSource);
        try {
            const output = execFileSync(
                this.$configuration.context.cmdAWS,
                [
                    "--output",
                    "json",
                    "s3",
                    (configuration.stateFileMethod == "move" ? "mv" : "cp"),
                    "s3://" + configuration.stateBucketName + "/" + configuration.stateFileSource,
                    "s3://" + configuration.stateBucketName + "/" + configuration.stateFileTarget
                ],
                {
                    timeout: 30 * 60 * 1000,
                    cwd: module.paths.directory,
                    env: _.merge({}, context.env || {}, process.env, module.env.toHash(), {
                        AWS_ACCESS_KEY_ID: configuration.stateCredentialsAccessKey,
                        AWS_SECRET_ACCESS_KEY: configuration.stateCredentialsSecretKey,
                        AWS_DEFAULT_REGION: configuration.stateRegion
                    })
                }
            );
            this.$cli.logInfo("State file moved!", output.toString());
        } catch (error) {
            throw new TaskExecutionException(
                "Unable to move state file!",
                configuration,
                error
            );
        }
        return true;
    }

    private async moveStateResources(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        const configuration = TerraformTaskMoveStateResourcesConfiguration.parse(
            task.configuration,
            module,
            binding
        );
        this.$cli.logNamedYamlObject(configuration, `Task '${task.name}' configuration`);
        if (!configuration.stateResourcesMove.size) {
            this.$cli.logDebug("No resources to move specified!");
            return false;
        }
        const stateExists = await this.hasBucketFile({
            path: configuration.stateDirectory + "/" + configuration.stateFile,
            bucket: {
                name: configuration.stateBucketName,
                region: configuration.stateRegion,
                credentials: {
                    accessKey: configuration.stateCredentialsAccessKey,
                    secretKey: configuration.stateCredentialsSecretKey
                }
            }
        }, module);
        if (!stateExists) {
            this.$cli.logWarn("Source state file is not found! Fresh deployment?", {
                path: configuration.stateDirectory + "/" + configuration.stateFile
            });
            return false;
        }
        this.$cli.logInfo("Moving terragrunt state resources...");
        const moveResources = Array.from(configuration.stateResourcesMove.entries());
        const stateResources = await this.listStateResources({}, context, module, binding);
        for (let index = 0; index < moveResources.length; index++) {
            const from = moveResources[index][0];
            const to = moveResources[index][1];
            if (stateResources.has(to)) {
                this.$cli.logWarn(`Resource '${from} –> ${to}' was already migrated!`);
                continue;
            }
            if (!stateResources.has(from)) {
                this.$cli.logWarn(`Unable to move resource '${from}' –> '${to}', resource not found!\nResources in state:`, Array.from(stateResources));
                continue;
            }
            try {
                await this.moveStateResource({from, to}, context, module, binding);
            } catch (error) {
                throw new TaskExecutionException("Error while moving resource!", {
                    configuration,
                    stateResources,
                    moveResources,
                }, error);
            }
        }
        return true;
    }

    private async pruneStateResources(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        const configuration = TerraformTaskPruneStateResourcesConfiguration.parse(
            task.configuration,
            module,
            binding
        );
        this.$cli.logNamedYamlObject(configuration, `Task '${task.name}' configuration`);
        const stateExists = await this.hasBucketFile({
            path: configuration.stateDirectory + "/" + configuration.stateFile,
            bucket: {
                name: configuration.stateBucketName,
                region: configuration.stateRegion,
                credentials: {
                    accessKey: configuration.stateCredentialsAccessKey,
                    secretKey: configuration.stateCredentialsSecretKey
                }
            }
        }, module);
        if (!stateExists) {
            this.$cli.logDebug("State file not found! Nothing to do...");
            return false;
        }
        const stateResources = await this.listStateResources({}, context, module, binding);
        const pruneResources = Array.from(stateResources).filter((resource) => {
            if (configuration.stateResourcesKeep.has(resource)
                || resource.startsWith("terraform_remote_state.")
                || Array.from(configuration.stateResourcesKeep.values()).find((keep) => resource.startsWith(keep))
            ) {
                this.$cli.logDebug(`Resource ${resource} will be kept in state!`);
                return false;
            }
            return true;
        });
        if (!pruneResources.length) {
            this.$cli.logInfo("Nothing to prune!\nResources left in state:", Array.from(stateResources));
            return false;
        }
        this.$cli.logInfo("Pruning terragrunt state resources:", pruneResources);
        for (let index = 0; index < pruneResources.length; index++) {
            const resource = pruneResources[index];
            try {
                await this.removeStateResource({resource}, context, module, binding, {
                    progress: `${index + 1}/${pruneResources.length}`
                });
            } catch (error) {
                new TaskExecutionException("Unable to remove resource!", {
                    configuration,
                    stateResources: Array.from(stateResources),
                    pruneResources: Array.from(pruneResources),
                }, error);
            }
        }
        const stateResourcesAfter = await this.listStateResources({}, context, module, binding);
        this.$cli.logInfo("Prune completed!\nResources left in state:", Array.from(stateResourcesAfter));
        return true;
    }

    private async listStateResources(
        payload: Primitives.Serialized,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ): Promise<Set<string>> {
        let trigger = false;
        const output = new Array<string>();
        await this.executeTerragrunt(
            "list state",
            ["state", "list"],
            context,
            module,
            binding,
            30,
            [
                {
                    action: async (data) => {
                        if (data.includes("terraform state list")) {
                            trigger = true;
                            return false;
                        }
                        if (trigger) {
                            output
                                .push(...data.split("\n")
                                    .map((item) => item.trim())
                                    .filter((item) => item.length));
                            return true;
                        }
                        return false;
                    }
                }
            ]
        );
        return new Set<string>(output);
    }

    private async moveStateResource(
        payload: Primitives.Serialized,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        const from = payload["from"];
        const to = payload["to"];
        this.$cli.logInfo(`Moving resource: '${from}' –> '${to}'`);
        await this.executeTerragrunt(
            `move resource '${from}' –> '${to}'`,
            ["state", "mv", from, to],
            context,
            module,
            binding,
            30,
            [
                {
                    action: async (data, event) => {
                        if (!this.$cli.quiet && !this.$cli.levels.debug) {
                            event.spinner.stop();
                            event.spinner.clear();
                        }
                        return false;
                    }
                }
            ]
        );
    }

    private async removeStateResource(
        payload: Primitives.Serialized,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module,
        metadata: Primitives.Serialized = {}
    ) {
        const resource = payload["resource"];
        if (metadata.progress) {
            this.$cli.logWarn(`[${metadata.progress}] Removing terragrunt state resource '${resource}'...`);
        } else {
            this.$cli.logWarn(`Removing terragrunt state resource '${resource}'...`);
        }
        await this.executeTerragrunt(
            `removing resource '${resource}'`,
            ["state", "rm", resource],
            context,
            module,
            binding,
            30
        );
    }

    private async executeTerragrunt(
        action: string,
        args: string[],
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module,
        timeout: number,
        patterns: Pattern[] = []
    ) {
        this.$cli.logDebug(`Executing terraform ${action}...`);
        const options = {
            timeout: timeout * 60 * 1000,
            cwd: module.paths.directory,
            env: _.merge({}, {}, context.env || {}, process.env, module.env.toHash(), {
                "AWS_ACCESS_KEY_ID": module.context.environment.terraform.backend.s3.credentials["access-key"],
                "AWS_SECRET_ACCESS_KEY": module.context.environment.terraform.backend.s3.credentials["secret-key"]
            })
        };
        this.$cli.logTrace("Executing terragrunt", {
            args: args,
            options: options
        });
        const childProcess = execFile(
            this.$configuration.context.cmdTerragrunt,
            args,
            options
        );
        const log: string[] = [];
        childProcess.stderr.on("data", (data) => log.push(data.trim()));
        childProcess.stdout.on("data", (data) => log.push(data.trim()));
        const spinner = ora.default({spinner: "bouncingBar", isEnabled: !this.$cli.quiet});
        try {
            await new Promise((resolve, reject) => {
                patterns.unshift(
                    {
                        action: async (data) => {
                            if (this.$cli.levels.debug) {
                                //this.$cli.logProcessOutput(data);
                            }
                            return false;
                        }
                    },
                    {
                        selector: "Terraform will perform the following",
                        action: async () => {
                            if (!this.$cli.levels.debug && !this.$cli.quiet) {
                                spinner.stop();
                                spinner.clear();
                            }
                            return false;
                        }
                    },
                    {
                        selector: "No changes. Infrastructure is up-to-date",
                        action: async () => {
                            if (!this.$cli.levels.debug && !this.$cli.quiet) {
                                spinner.stop();
                                spinner.clear();
                            }
                            this.$cli.logWarn("No changes detected!");
                            return true;
                        }
                    },
                    {
                        selector: "Refreshing Terraform state in-memory prior to plan",
                        action: async () => {
                            spinner.text = "Collecting resources state...";
                            return false;
                        }
                    },
                    {
                        selector: "Acquiring state lock.",
                        action: async () => {
                            spinner.text = "Acquiring state lock...";
                            return false;
                        }
                    },
                    {
                        selector: ["Error:", "Error loading state:", "Failed to load state:", "Item to move doesn't exist:", "Error applying plan:", "Hit multiple errors:"],
                        action: async (data, event) => {
                            if (!this.$cli.levels.debug && !this.$cli.quiet) {
                                spinner.stop();
                                spinner.clear();
                            }
                            //reject(new Exception(data.trim()));
                            return true;
                        }
                    },
                    {
                        selector: ["Do you want to copy existing state to the new backend?", "does not exist or you don't have permissions to access it"],
                        action: async (data, event) => {
                            if (!this.$cli.levels.debug && !this.$cli.quiet) {
                                spinner.stop();
                                spinner.clear();
                            }
                            this.$cli.logError("Corrupted backend configuration! Please check state in DynamoDB and clear terragrunt cache folders!");
                            process.exit(1);
                            return true;
                        }
                    },
                    {
                        action: async (data) => {
                            if (!this.$cli.levels.debug) {
                                spinner.text = data.substr(0, 120);
                            }
                            return false;
                        }
                    }
                );
                if (!this.$cli.levels.debug && !this.$cli.quiet) {
                    spinner.start(`Launching terraform ${action}...`);
                }
                ["stderr", "stdout"].forEach((channelName) => {
                    const channel = channelName == "stderr" ?
                        childProcess.stderr :
                        childProcess.stdout;
                    channel.on("data", async (lines) => {

                        lines = lines.split("\n")
                            .map((line: string) => line.trim())
                            .filter((line: string) => line.length);

                        for (let index = 0; index < lines.length; index++) {
                            const line = lines[index];

                            await (async function () {
                                for (let patternIndex = 0; patternIndex < patterns.length; patternIndex++) {
                                    const pattern = patterns[patternIndex];
                                    let selectors = (!(pattern.selector instanceof Array)) ? [pattern.selector] : pattern.selector;
                                    for (let selectorIndex = 0; selectorIndex < selectors.length; selectorIndex++) {
                                        const selector = selectors[selectorIndex];
                                        let match = false;
                                        if (selector == undefined) {
                                            match = true;
                                        } else if (typeof selector == "string") {
                                            match = line.includes(selector);
                                        }
                                        if (match) {
                                            if (await pattern.action(line, {
                                                channelName: channelName,
                                                spinner: spinner
                                            })) {
                                                return;
                                            }
                                        }
                                    }
                                }
                            }());


                        }
                    });
                });
                childProcess.addListener("error", reject);
                childProcess.addListener("exit", (exitCode) => {
                    if (!this.$cli.levels.debug && !this.$cli.quiet) {
                        spinner.stop();
                        spinner.clear();
                    }
                    if (exitCode) {
                        this.$cli.logProcessOutput(log.join("\n"));
                        reject(new Exception(`Terraform exited with '${exitCode}' code!`));
                    } else {
                        resolve();
                    }
                });
            });
            if (!this.$cli.levels.debug && !this.$cli.quiet) {
                spinner.stop();
                spinner.clear();
            }
        } catch (error) {
            throw new TaskExecutionException(`Terraform ${action} failed!`, {
                log: log.join("\n")
            }, error);
        }
    }

    private async plan(
        payload: Primitives.Serialized,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        let output = false;
        await this.executeTerragrunt(
            "plan",
            ["plan", "-out", module.paths.directory + "/~~terraform.plan"],
            context,
            module,
            binding,
            30,
            [
                {
                    selector: "------------------------------------------------------------------------",
                    action: async () => {
                        if (output) {
                            output = false;
                            return true;
                        }
                        return false;
                    }
                },
                {
                    action: async (data) => {
                        if (output) {
                            this.$cli.logProcessOutput(`\n\n\n${data.trim()}\n`);
                            return true;
                        } else {
                            return false;
                        }

                    },
                },
                {
                    selector: "Terraform will perform the following",
                    action: async () => {
                        output = true;
                        return false;
                    }
                }
            ]);
        return true;
    }

    private async output(
        payload: Primitives.Serialized,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        let output = false;
        await this.executeTerragrunt(
            "outp" +
            "ut", ["output", "-json"],
            context,
            module,
            binding,
            30,
            [
                {
                    selector: "terraform output -json",
                    action: async () => {
                        output = true;
                        return true;
                    }
                },
                {
                    action: async (data) => {
                        if (output) {
                            console.log(`\n${data.trim()}\n`);
                            return true;
                        } else {
                            return false;
                        }
                    }
                }
            ]);
        return true;
    }

    private async validate(
        payload: Primitives.Serialized,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        let output = false;
        await this.executeTerragrunt(
            "validate",
            ["validate"],
            context,
            module,
            binding,
            30,
            [
                {
                    action: async (data) => {
                        if (output) {
                            this.$cli.logProcessOutput(`\n${data.trim()}\n`);
                            return true;
                        } else {
                            return false;
                        }
                    }
                },
                {
                    selector: "-------------------------",
                    action: async () => {
                        output = false;
                        return false;
                    }
                }
            ]);
    }

    private async destroy(
        payload: Primitives.Serialized,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        let output = false;
        await this.executeTerragrunt(
            "destroy",
            ["destroy", "-auto-approve", "--terragrunt-non-interactive"],
            context,
            module,
            binding,
            30,
            [
                {
                    action: async (data) => {
                        if (output) {
                            this.$cli.logProcessOutput(`\n${data.trim()}\n`);
                            return true;
                        } else {
                            return false;
                        }
                    }
                },
                {
                    selector: ["terraform destroy"],
                    action: async (data, event) => {
                        event.spinner.stop();
                        event.spinner.clear();
                        output = true;
                        return false;
                    }
                },
                {
                    selector: ["Terraform will perform the following", "Destroy complete!", "Outputs:", "Refreshing state..."],
                    action: async (data, event) => {
                        event.spinner.stop();
                        event.spinner.clear();
                        if (!output) {
                            output = true;
                            this.$cli.logProcessOutput(`\n${data.trim()}\n`);
                            return false;
                        } else {
                            return true;
                        }
                    }
                }
            ]);
        return true;
    }

    private async deploy(
        payload: Primitives.Serialized,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        let output = false;
        const args = [
            "apply",
            "-auto-approve",
            "--terragrunt-non-interactive"
        ];
        if (fs.existsSync(module.paths.directory + "/~~terraform.plan")) {
            args.push(module.paths.directory + "/~~terraform.plan");
        }
        try {
            await this.executeTerragrunt(
                "deploy",
                args,
                context,
                module,
                binding,
                30,
                [
                    {
                        action: async (data) => {
                            if (output) {
                                this.$cli.logProcessOutput(`\n${data.trim()}\n`);
                                return true;
                            } else {
                                return false;
                            }
                        }
                    },
                    {
                        selector: ["terraform apply"],
                        action: async (data, event) => {
                            event.spinner.stop();
                            event.spinner.clear();
                            output = true;
                            return false;
                        }
                    },
                    {
                        selector: ["Terraform will perform the following", "Apply complete!", "Outputs:", "Refreshing state..."],
                        action: async (data, event) => {
                            event.spinner.stop();
                            event.spinner.clear();
                            if (!output) {
                                output = true;
                                this.$cli.logProcessOutput(`\n${data.trim()}\n`);
                                return false;
                            } else {
                                return true;
                            }
                        }
                    }
                ]);
            return true;
        } catch (error) {
            fs.writeFileSync(module.paths.directory + "/~~terraform.error", error.message);
            if (!fs.existsSync(module.paths.directory + "/on-terraform-deploy-error.sh")) {
                throw error;
            }
            this.$cli.logWarn("Terraform error caught, running recovery script!", error);
            execFileSync(
                "/usr/bin/env",
                ["bash", module.paths.directory + "/on-terraform-deploy-error.sh"],
                {
                    cwd: module.paths.directory,
                    encoding: "utf-8",
                    stdio: "inherit",
                    env: _.merge({}, context.env || {}, process.env, module.env.toHash())
                }
            );
            return -1;
        } finally {
            if (fs.existsSync(module.paths.directory + "/~~terraform.plan")) {
                fs.unlinkSync(module.paths.directory + "/~~terraform.plan");
            }
        }
    }

}
