import * as Primitives from "../../../../domain/Primitives";
import * as Models from "../../../../domain/Models";
import * as _ from "lodash";
import Exception from "../../../error/Exception";

const Validator = require("fastest-validator");

export default class TerraformTaskMoveStateResourcesConfiguration {
    constructor(
        readonly stateResourcesMove: Map<string, string>,
        readonly stateFile: string,
        readonly stateDirectory: string,
        readonly stateBucketName: string,
        readonly stateRegion: string,
        readonly stateCredentialsAccessKey: string,
        readonly stateCredentialsSecretKey: string
    ) {

    }

    static parse(
        configuration: Primitives.Serialized,
        module: Models.Module,
        binding: Models.Module
    ): TerraformTaskMoveStateResourcesConfiguration {
        if (!configuration) throw new Error("Missing configuration!");
        configuration = _.merge({
            state: {
                file: module.context.environment.terraform.state.id + ".tfstate",
                directory: module.context.environment.terraform.state.directory,
                bucket: {
                    name: module.context.environment.terraform.backend.s3.bucket.name,
                    region: module.context.environment.terraform.backend.s3.region,
                    credentials: {
                        "access-key": module.context.environment.terraform.backend.s3.credentials["access-key"],
                        "secret-key": module.context.environment.terraform.backend.s3.credentials["secret-key"]
                    }
                }
            }
        }, configuration);
        const validator: any = new Validator();
        const errors = validator.validate(configuration, {
            state: {
                type: "object",
                props: {
                    file: {type: "string"},
                    directory: {type: "string"},
                    bucket: {
                        type: "object",
                        props: {
                            name: {type: "string"},
                            region: {type: "string"},
                            credentials: {
                                type: "object",
                                props: {
                                    "access-key": {type: "string"},
                                    "secret-key": {type: "string"}
                                }
                            }
                        }
                    }
                }
            }
        } as any);
        if (errors !== true) throw new Exception("Invalid task configuration:", errors);
        const stateResourcesMove = new Map<string, string>();
        Object.keys(configuration.state.resources.move).forEach((oldPath) => {
            const newPath = configuration.state.resources.move[oldPath];
            stateResourcesMove.set(oldPath, newPath);
        });
        return new TerraformTaskMoveStateResourcesConfiguration(
            stateResourcesMove,
            configuration.state.file,
            configuration.state.directory,
            configuration.state.bucket.name,
            configuration.state.bucket.region,
            configuration.state.bucket.credentials["access-key"],
            configuration.state.bucket.credentials["secret-key"]
        );
    }
}
