import TaskExecutor from "../../TaskExecutor";
import {inject, injectable} from "inversify";
import Cli from "../../../cli/Cli";
import * as Models from "../../../../domain/Models";
import * as glob from "glob";
import * as _ from "lodash";
import * as fs from "fs";
import * as ejs from "ejs";
import * as Paths from "path"
import * as Primitives from "../../../../domain/Primitives";
import YamlService from "../../../yaml/YamlService";
import TaskExecutorContext from "../../TaskExecutorContext";
import YamlLazyValue from "../../../yaml/YamlLazyValue";

@injectable()
export default class TerraformTemplateSourcesGeneratorTaskExecutor
    implements TaskExecutor {

    public readonly name: string = "terraform-template-sources-generator";

    constructor(
        @inject(Cli)
        protected $cli: Cli,
        @inject(YamlService)
        private readonly $yaml: YamlService
    ) {
    }

    async execute(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        const cwd = module.paths.directory;
        const spiffFiles = glob.sync("*.tftemplate", {
            cwd: cwd
        });
        for (let index = 0; index < spiffFiles.length; index++) {
            const file = spiffFiles[index];
            await this.generateSpiffTemplate(
                fs.readFileSync(`${cwd}/${file}`).toString(),
                Paths.basename(file, ".tftemplate"),
                module,
                binding
            );
        }
        if (spiffFiles.length) {
            this.$cli.logInfo(`Generated ${spiffFiles.length} terraform files from spiff template!`);
        }
        const jsFiles = glob.sync("*.tfjs", {
            cwd: cwd
        });
        for (let index = 0; index < jsFiles.length; index++) {
            const file = jsFiles[index];
            await this.generateJsTemplate(
                fs.readFileSync(`${cwd}/${file}`).toString(),
                Paths.basename(file, ".tfjs"),
                module,
                binding
            );
        }
        if (jsFiles.length) {
            this.$cli.logInfo(`Generated ${jsFiles.length} terraform files from js template!`);
        }
        const ejsFiles = glob.sync("*.tfejs", {
            cwd: cwd
        });
        for (let index = 0; index < ejsFiles.length; index++) {
            const file = ejsFiles[index];
            await this.generateEjsTemplate(
                fs.readFileSync(`${cwd}/${file}`).toString(),
                Paths.basename(file, ".tfejs"),
                module,
                binding
            );
        }
        if (ejsFiles.length) {
            this.$cli.logInfo(`Generated ${ejsFiles.length} terraform files from ejs template!`);
        }
    }

    private async generateJsTemplate(source: string, name: string, module: Models.Module, binding: Models.Module) {
        this.$cli.logDebug(`Interpolating ${module.id} ${name}:`);
        const cwd = module.paths.directory;
        const output: any = [];
        new Function(
            "output",
            "module",
            "binding",
            source
        )(output, module.serialize(), binding.serialize());
        fs.writeFileSync(`${cwd}/~${name}.tf`, output.join("\n"));
        this.$cli.logDebug(`Generated ${cwd}/~${name}.tf terraform file from js template!`);
    }

    private async generateEjsTemplate(
        source: string,
        name: string,
        module: Models.Module,
        binding: Models.Module
    ) {
        this.$cli.logDebug(`Interpolating ${module.id} ${name}:`);
        const cwd = module.paths.directory;
        const output = ejs.render(source, {
            module: module.serialize(),
            binding: binding.serializeAsBinding(),
            require: require,
            application: (global as any)['application']
        });
        fs.writeFileSync(`${cwd}/~${name}.tf`, output);
        this.$cli.logDebug(`Generated ${cwd}/~${name}.tf terraform file from js template!`);
    }

    private async generateSpiffTemplate(source: string, name: string, module: Models.Module, binding: Models.Module) {
        this.$cli.logDebug(`Interpolating ${module.id} ${name}:`);
        const cwd = module.paths.directory;
        const spiffExpressions = new Array<string>()
            .concat(source.match(/(?<=<<SPIFF)(.*?)(?=SPIFF)/gmus) || [])
            .concat(source.match(/(?<=<<\()(.*?)(?=\)>>)/gmus) || []);
        if (!spiffExpressions.length) {
            return;
        }
        const inputStructure: Primitives.Serialized = {};
        spiffExpressions.forEach((expression, index) => {
            inputStructure["exp" + index] = expression.trim();
        });
        const inputSource = this.$yaml.stringify(inputStructure);
        const inputFile = `${cwd}/~~${name}.input.yaml`;
        fs.writeFileSync(inputFile, inputSource);
        const variablesData: Primitives.Serialized = {
            module: module.serialize()
        };
        const variablesStructure: Primitives.Serialized = {
            module: "(( &temporary ))"
        };
        if (binding) {
            variablesData.binding = binding.serialize();
            variablesStructure.binding = "(( &temporary ))";
        }
        const variablesYaml = this.$yaml.stringify(variablesData);
        const variablesFile = `${cwd}/~~${name}.variables.yaml`;
        const templateData = _.merge(
            this.$yaml.parseFile(inputFile),
            variablesStructure
        );
        const templateYaml = this.$yaml.stringify(templateData);
        const templateFile = `${cwd}/~~${name}.template.yaml`;
        fs.writeFileSync(templateFile, templateYaml);
        fs.writeFileSync(variablesFile, variablesYaml);
        const resultYaml = this.$yaml.spiffFiles([templateFile, variablesFile], {
            env: module.env.filter((value) => !(value instanceof YamlLazyValue)).serialize(),
            cwd: cwd
        });
        const resultStructure = this.$yaml.parse(
            resultYaml,
            {
                env: module.env.filter((value) => !(value instanceof YamlLazyValue)).serialize(),
                cwd: cwd
            }
        ) as any;
        const resultFile = `${cwd}/~~${name}.result.yaml`;
        fs.writeFileSync(resultFile, resultYaml);
        spiffExpressions.forEach((expression, index) => {
            source = source.replace("<<SPIFF" + expression + "SPIFF\n", JSON.stringify(typeof resultStructure["exp" + index] == "string" ?
                resultStructure["exp" + index].trim() :
                resultStructure["exp" + index]
            ));
            source = source.replace("<<(" + expression + ")>>", resultStructure["exp" + index]);
        });
        fs.writeFileSync(`${cwd}/~${name}.tf`, source);
        this.$cli.logDebug(`Generated ${cwd}/~${name}.tf terraform file from spiff template!`);
    }

}
