import * as fs from "fs";
import * as _ from "lodash";
import * as flat from "flat";
import TerraformVariablesGeneratorTaskConfiguration, {ValueFormat} from "./TerraformVariablesGeneratorTaskConfiguration";
import TaskExecutor from "../../TaskExecutor";
import {inject, injectable} from "inversify";
import Cli from "../../../cli/Cli";
import TaskExecutionError from "../../TaskExecutionError";
import * as Models from "../../../../domain/Models";
import YamlService from "../../../yaml/YamlService";
import TaskExecutorContext from "../../TaskExecutorContext";
import YamlLazyValue from "../../../yaml/YamlLazyValue";
import * as glob from "glob";

@injectable()
export default class TerraformVariablesGeneratorTaskExecutor
    implements TaskExecutor {

    public readonly name: string = "terraform-variables-generator";

    constructor(
        @inject(Cli)
        protected $cli: Cli,
        @inject(YamlService)
        private readonly $yaml: YamlService
    ) {

    }

    async execute(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        if (!task.configuration) throw new Error;
        const configuration = new TerraformVariablesGeneratorTaskConfiguration(
            task.configuration["output-file-prefix"],
            task.configuration["template-section"],
            task.configuration["definitions-create"],
            task.configuration["values-format"],
            task.configuration["defined-only"],
            task.configuration["context"]
        );
        const cwd = module.paths.directory;
        let inputFile = `${cwd}/terraform.tfvars.yaml`;
        if (!fs.existsSync(inputFile) && module.templates.has(configuration.templateSection)) {
            inputFile = `${cwd}/~~terraform.variables.input.yaml`;
            let template = module.templates.get(configuration.templateSection);
            if (typeof template != "string") {
                template = this.$yaml.stringify(template);
            }
            let inputStructure = this.$yaml.parse(template, {
                cwd
            });
            if (inputStructure instanceof Array) {
                inputStructure = _.merge({}, ...inputStructure);
            }
            fs.writeFileSync(inputFile, this.$yaml.stringify(inputStructure));
        } else {
            if (task.metadata.enabled) {
                throw new TaskExecutionError(`No template '${configuration.templateSection}' defined! Found: ${JSON.stringify(module.templates.getNames())}!`);
            } else {
                this.$cli.logDebug(`No terraform.tfvars.yaml file found or '${configuration.templateSection}' template defined!`);
                return;
            }
        }
        const contextStructure = {
            module: module.serialize(),
            binding: binding ? binding.serializeAsBinding() : undefined
        };
        if (!contextStructure.binding) {
            delete contextStructure.binding;
        }
        const inputStructure = this.$yaml.parseFile(inputFile);
        const templateStructure = {
            variables: inputStructure,
            module: "(( &temporary ))",
            binding: "(( &temporary ))"
        };
        const templateYaml = this.$yaml.stringify(templateStructure);
        const templateFile = `${cwd}/~~terraform.variables.template.yaml`;
        const contextYaml = this.$yaml.stringify(contextStructure);
        const contextFile = `${cwd}/~~terraform.variables.context.yaml`;
        this.$cli.logDebug(`Interpolating ${module.id} ${configuration.templateSection}:`);
        fs.writeFileSync(templateFile, templateYaml);
        fs.writeFileSync(contextFile, contextYaml);
        const outputYaml = this.$yaml.spiffFiles([templateFile, contextFile], {
            env: module.env.filter((value) => !(value instanceof YamlLazyValue)).serialize(),
            cwd: cwd
        });
        const outputStructure = this.$yaml.parse(
            outputYaml,
            {
                env: module.env.filter((value) => !(value instanceof YamlLazyValue)).serialize(),
                cwd: cwd
            }
        ) as any;
        const variablesStructure: { [key: string]: Object } =
            flat(outputStructure.variables, {delimiter: "-", safe: true});
        const maxDepth = Object.keys(variablesStructure)
            .map((s) => s.split("-").length)
            .reduce((p, c, i, a) => Math.max((p || 0), c));
        for (let depth = 1; depth < (maxDepth - 1); depth++) {
            const depthStructure: { [key: string]: Object } = flat(outputStructure.variables, {delimiter: "-", safe: true, maxDepth: depth});
            for (let key in depthStructure) {
                if (typeof variablesStructure[key] == "undefined") {
                    variablesStructure[key] = depthStructure[key];
                }
            }
        }
        fs.writeFileSync(`${cwd}/~~terraform.variables.result.yaml`, outputYaml);
        if (configuration.valuesFormat == ValueFormat.JSON) {
            const variablesJson = JSON.stringify(variablesStructure, null, 4);
            fs.writeFileSync(`${cwd}/${configuration.outputFilePrefix}.tfvars.json`, variablesJson);
        }
        this.$cli.logNamedYaml(outputYaml, `Interpolation module:${module.id} manifest:${configuration.templateSection}:`);
        const variablesFilter: string[] = []
        if (configuration.definedOnly) {
            glob.sync("*.tf", {
                sync: true,
                absolute: true,
                realpath: true,
                cwd: cwd
            }).forEach((file) => {
                const content = fs.readFileSync(file).toString();
                (content.match(/var\.([a-zA-Z0-9_-]+)/gm) || []).forEach((definition) =>
                    variablesFilter.push(definition.substr(4, definition.length - 4))
                )
            })
        }
        if (configuration.definitionsCreate) {
            const definitionsCode = Object
                .keys(variablesStructure)
                .map((variable) => {
                    if (configuration.definedOnly && !variablesFilter.includes(variable)) {
                        return ""
                    }
                    const name = variable.replace(/([.\/])/g, "-");
                    const type = (
                        variablesStructure[variable] instanceof Array ? "list" :
                            variablesStructure[variable] instanceof Object ? "map" : "string"
                    );
                    const defaults = (function () {
                        if (configuration.valuesFormat != ValueFormat.DEFAULTS) {
                            return;
                        }
                        // Consider better replacer for objects
                        if (type != "string") {
                            return JSON.stringify(variablesStructure[variable], null, 4).replace(/": /g, "\" = ");
                        } else {
                            return JSON.stringify(variablesStructure[variable], null, 4);
                        }
                    }());
                    if (defaults == "null") return "";
                    const buffer = [`variable "${name}" {`, `\ttype = "${type}"`];
                    if (defaults != undefined) buffer.push(`\tdefault = ${defaults}`);
                    buffer.push(`}`);
                    return buffer.join("\n");
                })
                .filter((string) => string.length)
                .join("\n");
            fs.writeFileSync(
                `${cwd}/${configuration.outputFilePrefix}.tf`,
                definitionsCode
            );
            this.$cli.logInfo(`Generated terraform variables '${configuration.outputFilePrefix}.tf'`);
        }
    }

}
