import TaskExecutor from "../../TaskExecutor";
import TerragruntConfigurationGeneratorTaskConfiguration from "./TerragruntConfigurationGeneratorTaskConfiguration";
import {execFileSync} from "child_process";
import * as _ from "lodash";
import * as fs from "fs";
import {inject, injectable} from "inversify";
import Cli from "../../../cli/Cli";
import * as Models from "../../../../domain/Models";
import {Serialized} from "../../../../domain/Primitives";
import YamlService from "../../../yaml/YamlService";
import TaskExecutorContext from "../../TaskExecutorContext";
import YamlLazyValue from "../../../yaml/YamlLazyValue";

@injectable()
export default class TerragruntConfigurationGeneratorTaskExecutor
    implements TaskExecutor {

    public readonly name: string = "terragrunt-configuration-generator";

    constructor(
        @inject(Cli)
        protected $cli: Cli,
        @inject(YamlService)
        private readonly $yaml: YamlService
    ) {
    }

    async execute(
        task: Models.ModuleTask,
        context: TaskExecutorContext,
        module: Models.Module,
        binding: Models.Module
    ) {
        const configuration = TerragruntConfigurationGeneratorTaskConfiguration.parse(
            task.configuration,
            module,
            binding
        );
        this.$cli.logNamedYamlObject(configuration, `Task '${task.name}' configuration`);
        this.$cli.logDebug(`Interpolating ${module.id} ${configuration.templateSection}:`);
        const cwd = module.paths.directory;
        const inputHclFile = `${cwd}/~~terragrunt.configuration.input.hcl`;
        const inputJsonFile = `${cwd}/~~terragrunt.configuration.input.json`;
        const inputYamlFile = `${cwd}/~~terragrunt.configuration.input.yaml`;
        const inputHcl = module.templates.get(configuration.templateSection);
        fs.writeFileSync(inputHclFile, inputHcl);
        const inputJson = execFileSync(
            "bash",
            ['-c', `json2hcl -reverse < ${inputHclFile}`],
            {
                cwd: cwd,
                env: _.merge({}, process.env, module.env.toHash())
            }
        ).toString();
        fs.writeFileSync(inputJsonFile, inputJson);
        const inputStructure = JSON.parse(inputJson);
        const inputYaml = this.$yaml.stringify(inputStructure);
        fs.writeFileSync(inputYamlFile, inputYaml);
        //
        const variablesData: Serialized = {
            module: module.serialize()
        };
        const variableStructure: Serialized = {
            module: "(( &temporary ))"
        };
        if (binding) {
            variablesData.binding = binding.serialize();
            variableStructure.binding = "(( &temporary ))";
        }

        const templateYaml = this.$yaml.stringify(_.merge(variableStructure, inputStructure));
        const templateFile = `${cwd}/~~terragrunt.configuration.template.yaml`;
        fs.writeFileSync(templateFile, templateYaml);

        const variables = this.$yaml.stringify(variablesData);
        const variablesFile = `${cwd}/~~terragrunt.configuration.variables.yaml`;
        fs.writeFileSync(variablesFile, variables);
        try {
            const resultStructure = this.$yaml.parseSpiffFiles([templateFile, variablesFile], {
                env: module.env.filter((value) => !(value instanceof YamlLazyValue)).serialize(),
                cwd: cwd
            });
            const resultJson = JSON.stringify(resultStructure, null, 4);
            const jsonFile = `${cwd}/~~terragrunt.configuration.result.json`;
            fs.writeFileSync(jsonFile, resultJson);
            const resultHcl = execFileSync(
                "bash",
                ['-c', `json2hcl < ${jsonFile}`],
                {
                    cwd: cwd,
                    env: _.merge({}, process.env, module.env.toHash())
                }
            ).toString()
                .replace(/\n\n/g, "\n")
                .split("\n")
                .map((line) => line.replace(/^(\s*)?"([a-z_-]*)"/g, "$1$2"))
                .join("\n");
            fs.writeFileSync(`${cwd}/${configuration.outputFileName}`, resultHcl);
            this.$cli.logInfo(`Generated terragrunt configuration '${configuration.outputFileName}'`);
        } catch (error) {
            this.$cli.logNamedYamlObject(variablesData, "Context", true);
            throw error;
        }
    }

}
