import * as Path from "path";
import * as fs from "fs";
import Exception from "../error/Exception";
import * as yaml from "yaml";
import {ast} from "yaml";
import {injectable} from "inversify";
import {execFileSync} from "child_process";
import * as _ from "lodash";
import YamlLazyValue from "./YamlLazyValue";
import * as CronParser from "cron-parser";
import * as Moment from "moment";

@injectable()
export default class YamlService {

    constructor() {

    }

    public parseJsonFile(
        path: string,
        context: { cwd?: string } = {}
    ): Object {

        path = fs.realpathSync(path);

        if (!fs.existsSync(path)) {
            throw `Module manifest path ${path} doesn't exists!`;
        }

        if (!fs.lstatSync(path).isFile()) {
            throw `Module manifest path ${path} is not a file!`;
        }

        const source = fs.readFileSync(path).toString("utf-8");
        return JSON.parse(source);
    }

    public parseFile(
        path: string,
        context: {} = {}
    ): Object {

        path = fs.realpathSync(path);

        if (!fs.existsSync(path)) {
            throw `Module manifest path ${path} doesn't exists!`;
        }

        if (!fs.lstatSync(path).isFile()) {
            throw `Module manifest path ${path} is not a file!`;
        }

        const source = fs.readFileSync(path).toString("utf-8");
        return this.parse(source, _.merge({
            path: path
        }, context));
    }

    public stringify(source: Object, context: { path?: string } = {}): string {
        try {
            return JSON.stringify(source, null, 2)
        } catch (error) {
            throw new Exception(
                "Unable to stringify yaml",
                {
                    Source: source,
                    Path: context.path
                },
                error
            );
        }
    }

    public parseSpiffFiles(
        paths: string[],
        context: {} = {}
    ) {
        return this.parse(
            this.spiffFiles(paths, context),
            _.merge({
                path: "spiff " + paths.join(" ")
            }, context)
        );
    }

    public spiffFiles(paths: string | string[], context: { cwd?: string, env?: {}, timeout?: Number, asJson?: boolean } = {}) {
        const options = {
            cwd: context.cwd || Path.dirname(paths[0]),
            env: _.merge({}, process.env, context.env),
            encoding: "utf-8",
            timeout: 60000
        };
        const args = ["m"];
        if (context.asJson) {
            args.push("--json")
        }
        args.push(...paths);
        ((global as any).application as any).$cli.logTrace("Spiffing yaml files...", {
            command: "spiff++",
            args: args,
            options: options
        });
        return execFileSync(
            "spiff++",
            args,
            options
        ).toString("utf-8");
    }

    public spiffStream(content: string, context: { cwd?: string, env?: {}, timeout?: Number } = {}) {
        const options = {
            cwd: context.cwd,
            env: _.merge({}, process.env, context.env),
            encoding: "utf-8",
            timeout: 60000,
            input: content
        };
        ((global as any).application as any).$cli.logTrace("Spiffing yaml stream...", {
            command: "spiff++",
            args: ["m", "-"],
            options: options
        });
        return execFileSync(
            "spiff++",
            ["m", "-"],
            options
        ).toString("utf-8");
    }

    public parse(
        source: string,
        context: any = {}
    ): Object {
        try {
            source = source
                .replace(/<<: !/g, "<<-: !")
                .replace(/"<<": !/g, "<<-: !");
            return yaml.parse(source, {
                merge: false,
                keepNodeTypes: true,
                keepCstNodes: true,
                customTags: this.createTags(context)
            } as any);
        } catch (error) {
            throw new Exception(
                "Unable to parse yaml",
                {
                    Source: source,
                    Path: context.path || ""
                },
                error
            );
        }
    }

    private createTags(context: { path?: string, cwd?: string }) {
        if (context.path && !context.cwd) {
            context.cwd = Path.dirname(context.path);
        }
        return [
            {
                class: YamlLazyValue,
                default: false,
                tag: '!lazy',
                resolve: (document: any, node: any) => {
                    const value = node.context.src.substr(
                        node.valueRange.start,
                        node.valueRange.end - node.valueRange.start
                    );
                    if (!value.length) {
                        throw new Exception("Invalid !lazy value");
                    }
                    return new YamlLazyValue(value);
                },
                stringify: (item: ast.Node) => {
                    return (item as any).value.value.toString()
                }
            },
            {
                class: Function,
                default: false,
                tag: '!js',
                resolve: (document: any, node: any) => {
                    const body = node.context.src.substr(
                        node.valueRange.start,
                        node.valueRange.end - node.valueRange.start
                    );
                    try {
                        // @ts-ignore
                        (global as any)['context'] = context;
                        return eval(
                            "(" + body.trim() + ")"
                        );
                    } catch (error) {
                        throw new Exception(
                            "Invalid !js tag: " + error.message,
                            {
                                Source: body
                            },
                            error
                        );
                    }
                },
                stringify: (item: ast.Node) => {
                    return JSON.stringify((item as any).value.toString())
                }
            },
            {
                class: Function,
                default: false,
                tag: '!include.js',
                resolve: (document: any, node: any) => {
                    const body = node.context.src.substr(
                        node.valueRange.start,
                        node.valueRange.end - node.valueRange.start
                    );
                    try {
                        return require(context.cwd + "/" + body.trim())(context || {})
                    } catch (error) {
                        throw new Exception(
                            "Invalid !include.js tag: " + error.message,
                            {
                                Source: body
                            },
                            error
                        );
                    }
                },
                stringify: (item: ast.Node) => {
                    return JSON.stringify((item as any).value.toString())
                }
            },
            {
                class: String,
                default: false,
                tag: '!direct-inheritor',
                resolve: (document: any, node: any) => {
                    return '!direct-inheritor ' + node.context.src.substr(
                        node.valueRange.start,
                        node.valueRange.end - node.valueRange.start
                    );
                },
                stringify: () => {
                }
            },
            {
                class: String,
                default: false,
                tag: '!path.resolve',
                resolve: (document: any, node: any) => {
                    if (!context.cwd) {
                        throw new Exception("No current directory context!");
                    }
                    return Path.resolve(
                        context.cwd,
                        node.context.src.substr(
                            node.valueRange.start,
                            node.valueRange.end - node.valueRange.start
                        )
                    );
                },
                stringify: () => {
                }
            },
            {
                class: String,
                default: false,
                tag: '!include.file',
                resolve: (document: any, node: any) => {
                    if (!context.cwd) {
                        throw new Exception("No current directory context!");
                    }
                    const includePath = Path.resolve(
                        context.cwd,
                        node.context.src.substr(
                            node.valueRange.start,
                            node.valueRange.end - node.valueRange.start
                        )
                    );
                    return fs.readFileSync(includePath.trim()).toString()
                },
                stringify: () => {
                }
            } as any,
            {
                class: String,
                default: false,
                tag: '!include.yaml',
                resolve: (document: any, node: any) => {
                    if (!context.cwd) {
                        throw new Exception("No current directory context!");
                    }
                    const includePath = Path.resolve(
                        context.cwd,
                        node.context.src.substr(
                            node.valueRange.start,
                            node.valueRange.end - node.valueRange.start
                        )
                    ).trim();
                    return this.parseFile(includePath.trim());
                },
                stringify: () => {
                }
            } as any,
            {
                class: String,
                default: false,
                tag: '!env',
                resolve: (document: any, node: any) => {
                    const name = node.context.src.substr(
                        node.valueRange.start,
                        node.valueRange.end - node.valueRange.start
                    );
                    const variable = name.split(" ", 1)[0];
                    const defaults = name.split(" ").slice(1).join(" ");
                    const value = process.env.hasOwnProperty(variable) ?
                        process.env[variable] :
                        defaults;
                    if (typeof value == "undefined") {
                        throw new Exception(`Environment variable is not found!`, {
                            Name: name,
                            Defaults: defaults,
                            Env: process.env[variable]
                        });
                    }
                    return value;
                },
                stringify: () => {
                }
            } as any,
            {
                class: String,
                default: false,
                tag: '!cron.next',
                resolve: (document: any, node: any) => {
                    const pattern = node.context.src.substr(
                        node.valueRange.start,
                        node.valueRange.end - node.valueRange.start
                    );
                    return CronParser.parseExpression(JSON.parse(pattern), {utc: true}).next().toISOString();
                },
                stringify: (item: ast.Node) => {
                    return JSON.stringify((item as any).value.toString())
                }
            } as any,
            {
                class: String,
                default: false,
                tag: '!moment.add',
                resolve: (document: any, node: any) => {
                    const pattern = node.context.src.substr(
                        node.valueRange.start,
                        node.valueRange.end - node.valueRange.start
                    );
                    return Moment().add(JSON.parse(pattern)).toISOString();
                },
                stringify: (item: ast.Node) => {
                    return JSON.stringify((item as any).value.toString())
                }
            } as any,
            {
                class: String,
                default: false,
                tag: '!moment.add-start-of-day',
                resolve: (document: any, node: any) => {
                    const pattern = node.context.src.substr(
                        node.valueRange.start,
                        node.valueRange.end - node.valueRange.start
                    );
                    return Moment().add(JSON.parse(pattern)).startOf('day').toISOString();
                },
                stringify: (item: ast.Node) => {
                    return JSON.stringify((item as any).value.toString())
                }
            } as any
        ];
    }

}
