import {inject, injectable} from "inversify";
import * as _ from "lodash";
import * as Manifests from "./Manifests";
import * as Models from "./Models";
import SchemaValidator from "./SchemaValidator";
import ModuleMetadataInheritanceInto from "./models/ModuleMetadataInheritanceInto";
import YamlService from "../components/yaml/YamlService";
import traverse = require("traverse");
import YamlLazyValue from "../components/yaml/YamlLazyValue";

const packageJson = require("./../../package.json");
@injectable()
export default class ModelsBuilder {

    constructor(
        @inject(SchemaValidator)
        private readonly $validator: SchemaValidator,
        @inject(YamlService)
        private readonly $yaml: YamlService
    ) {

    }

    buildModule(manifest: Manifests.Manifest): Models.Module {
        this.$validator.validateManifest(manifest);
        const module = manifest.module;
        if (!module) {
            throw new SyntaxError(`Missing .module definition!`);
        }
        this.$validator.validateModule(module);
        const moduleMetadata = this.parseModuleMetadata(module, module.metadata);
        const moduleInputParameters = this.parseModuleInputParameters(module, module.inputs);
        const modulePaths = this.parseModulePaths(module, module.paths);
        const moduleEnv = this.parseModuleEnv(module, module.env);
        const moduleContext = this.parseModuleContext(module, module.context);
        const moduleLifecycles = this.parseModuleLifecycles(module, module.lifecycles);
        const moduleTasks = this.parseModuleTasks(module, module.tasks);
        const moduleTemplates = this.parseModuleTemplates(module, module.templates, moduleMetadata);
        const moduleHelpers = this.parseModuleHelpers(module, module.helpers);
        return new Models.Module(
            module.id,
            moduleMetadata,
            moduleInputParameters,
            modulePaths,
            moduleEnv,
            moduleLifecycles,
            moduleTasks,
            moduleTemplates,
            moduleHelpers,
            moduleContext,
        );
    }

    private parseModuleMetadata(
        module: Manifests.Module,
        metadata: Manifests.ModuleMetadata | undefined = {}
    ): Models.ModuleMetadata {
        if (!metadata) {
            throw new SyntaxError(`Missing .module.metadata section!`);
        }
        return new Models.ModuleMetadata(
            metadata.enabled == undefined ? true : metadata.enabled,
            metadata.virtual == undefined ? false : metadata.virtual,
            metadata.binding == undefined ? false : metadata.binding,
            this.parseModuleMetadataInheritance(module, metadata.inherit),
            this.parseModuleMetadataTags(module, metadata.tags),
            metadata.schema == undefined ? packageJson.version : metadata.schema,
            this.parseModuleMetadataAutoload(module, metadata.autoload),
            metadata.$interpolation || "none",
            metadata.$parents == undefined ? new Set() : new Set(metadata.$parents)
        );
    }

    private parseModuleMetadataInheritance(
        module: Manifests.Module,
        inherit: Manifests.ModuleMetadataInherit | boolean | undefined
    ): Models.ModuleMetadataInheritance {
        switch (true) {
            case !inherit:
                return new Models.ModuleMetadataInheritance();
            case inherit === true:
                return new Models.ModuleMetadataInheritance(
                    new ModuleMetadataInheritanceInto(
                        {
                            child: true
                        }
                    )
                );
            case typeof inherit == "object":
                inherit = inherit as Manifests.ModuleMetadataInherit;
                const inheritInto = !inherit.into ?
                    new ModuleMetadataInheritanceInto() :
                    new ModuleMetadataInheritanceInto(inherit.into);
                return new Models.ModuleMetadataInheritance(inheritInto);
            default:
                throw new SyntaxError(
                    "Invalid .module.metadata.inherit section! " +
                    JSON.stringify(inherit)
                );
        }
    }

    private parseModuleMetadataTags(
        module: Manifests.Module,
        tags: string[] | undefined
    ): Set<string> {
        const result = new Set<string>();
        if (tags) {
            tags.forEach((tag) => result.add(tag));
        }
        return result;
    }

    private parseModuleMetadataAutoload(
        module: Manifests.Module,
        patterns: string[] | undefined
    ): Set<string> {
        const result = new Set<string>();
        if (patterns) {
            patterns.forEach((pattern) => result.add(pattern));
        }
        return result;
    }

    private parseModulePaths(
        module: Manifests.Module,
        paths: Manifests.ModulePaths | undefined,
    ) {
        if (!paths) throw new SyntaxError("Missing .module.paths section!");
        if (!paths.manifest) throw new SyntaxError("Missing .module.paths.manifest field!");
        if (!paths.module) throw new SyntaxError("Missing .module.paths.module field!");
        if (!paths.directory) throw new SyntaxError("Missing .module.paths.directory field!");
        if (!paths.unit) throw new SyntaxError("Missing .module.paths.unit field!");
        if (!paths.root) throw new SyntaxError("Missing .module.paths.root field!");
        return new Models.ModulePaths(
            paths.manifest,
            paths.module,
            paths.directory,
            paths.unit,
            paths.root
        );
    }

    private parseModuleEnv(
        module: Manifests.Module,
        env: { [name: string]: string } | undefined
    ) {
        if (!env) {
            return new Models.ModuleEnv();
        }
        if (typeof env == "object") {
            const map = new Map<string, string>();
            Object.keys(env).forEach((key) => map.set(key, env[key]));
            return new Models.ModuleEnv(map);
        }
        throw new SyntaxError("Invalid .module.env section!");
    }

    private parseModuleContext(
        module: Manifests.Module,
        context: { [key: string]: any } | undefined
    ) {
        return context || {};
    }

    private parseModuleLifecycles(
        module: Manifests.Module,
        lifecycles: { [name: string]: Manifests.ModuleLifecycle } | Manifests.ModuleLifecycle[] | undefined
    ) {
        if (!lifecycles) {
            return new Models.ModuleLifecycles();
        }
        if (lifecycles instanceof Array) {
            return this.parseModuleLifecyclesArray(module, lifecycles);
        }
        if (lifecycles instanceof Object) {
            return this.parseModuleLifecyclesHash(module, lifecycles);
        }
        throw new SyntaxError("Invalid .module.lifecycles definition: not a object or array!");
    }

    private parseModuleInputParameters(
        module: Manifests.Module,
        inputs: Manifests.ModuleInputParameters | undefined
    ) {
        if (!inputs) {
            return new Models.ModuleInputParameters();
        }
        if (inputs instanceof Object) {
            return this.parseModuleInputParametersHash(module, inputs);
        }
        throw new SyntaxError("Invalid .module.inputs definition: not a object!");
    }

    private parseModuleHelpers(
        module: Manifests.Module,
        helpers: Manifests.ModuleHelpers | undefined
    ) {
        if (!helpers) {
            return new Models.ModuleHelpers();
        }
        if (helpers instanceof Object) {
            return this.parseModuleHelpersHash(module, helpers);
        }
        throw new SyntaxError("Invalid .module.helpers definition: not a object!");
    }

    private parseModuleTasks(
        module: Manifests.Module,
        tasks: { [name: string]: Manifests.ModuleTask } | Manifests.ModuleTask[] | undefined
    ) {

        if (!tasks) {
            return new Models.ModuleTasks();
        }
        if (tasks instanceof Array) {
            return this.parseModuleTasksArray(module, tasks);
        }
        if (tasks instanceof Object) {
            return this.parseModuleTasksHash(module, tasks);
        }
        throw new SyntaxError("Invalid .module.tasks definition: not a object or array!");

    }

    private parseModuleTemplates(
        module: Manifests.Module,
        templates: { [name: string]: string | any } | undefined,
        metadata: Models.ModuleMetadata
    ) {
        const collection = new Map<string, string | object>();
        if (templates) {
            Object.keys(templates).forEach((name) => {
                let template = templates[name];
                collection.set(name, template);
            });
        }
        return new Models.ModuleTemplates(collection);
    }

    private parseModuleLifecyclesArray(
        module: Manifests.Module,
        lifecycles: Manifests.ModuleLifecycle[]
    ) {
        const collection = new Map<string, Models.ModuleLifecycle>();
        lifecycles.forEach((lifecycle, index) => {
            if (!lifecycle.name) throw new SyntaxError;
            collection.set(lifecycle.name, this.parseModuleLifecycle(module, lifecycle))
        });
        return new Models.ModuleLifecycles(collection);
    }

    private parseModuleLifecyclesHash(
        module: Manifests.Module,
        lifecycles: { [name: string]: Manifests.ModuleLifecycle }
    ) {
        const collection = new Map<string, Models.ModuleLifecycle>();
        for (const name in lifecycles) {
            const lifecycle = lifecycles[name];
            if (!(lifecycle instanceof Object))
                throw new SyntaxError("Invalid lifecycle definition: not a object!");
            collection.set(
                name,
                this.parseModuleLifecycle(
                    module,
                    _.merge({}, lifecycles[name], {name: name}),
                )
            );
        }
        return new Models.ModuleLifecycles(collection);
    }

    private parseModuleLifecycle(
        module: Manifests.Module,
        lifecycle: Manifests.ModuleLifecycle
    ) {
        if (!lifecycle.name) {
            throw new SyntaxError;
        }
        return new Models.ModuleLifecycle(
            lifecycle.name,
            lifecycle.dependencies || [],
            this.parseModuleLifecycleActions(module, lifecycle.actions),
            this.parseModuleLifecycleActions(module, lifecycle["actions-before"]),
            this.parseModuleLifecycleActions(module, lifecycle["actions-after"])
        );
    }

    private parseModuleLifecycleActions(
        module: Manifests.Module,
        actions: Array<Manifests.ModuleLifecycleAction | string> | undefined,
    ) {
        const collection = new Array<Models.ModuleLifecycleAction>();
        if (actions) {
            if (!(actions instanceof Array)) {
                throw new SyntaxError("Invalid lifecycle tasks definition: not array!");
            }
            collection.push(...actions.map((action) =>
                this.parseModuleLifecycleAction(module, action)
            ));
        }
        return new Models.ModuleLifecycleActions(collection)
    }

    private parseModuleLifecycleAction(
        module: Manifests.Module,
        action: Manifests.ModuleLifecycleAction | string,
    ) {
        if (typeof action == "string")
            return new Models.ModuleLifecycleAction(
                action,
                new Models.ModuleLifecycleActionTask(
                    action,
                    {}
                )
            )

        if (action instanceof Object) {
            if (!action.task || typeof action.task == "string") {
                if (!action.name) {
                    throw new SyntaxError;
                }
                if (!action.task) {
                    throw new SyntaxError;
                }
                return new Models.ModuleLifecycleAction(
                    action.name,
                    this.parseModuleLifecycleActionTask(
                        module,
                        {name: action.task || action.name},
                    )
                );
            } else {
                if (!action.name) throw new SyntaxError;
                return new Models.ModuleLifecycleAction(
                    action.name,
                    this.parseModuleLifecycleActionTask(module, action.task)
                );
            }
        } else {
            throw new SyntaxError("Invalid lifecycle action definition!")
        }
    }

    private parseModuleLifecycleActionTask(
        module: Manifests.Module,
        task: Manifests.ModuleLifecycleActionTask
    ) {
        if (!task.name) {
            throw new SyntaxError();
        }
        return new Models.ModuleLifecycleActionTask(
            task.name,
            task.configuration
        )
    }

    private parseModuleTasksArray(
        module: Manifests.Module,
        tasks: Manifests.ModuleLifecycleActionTask[]
    ) {
        const collection = new Map<string, Models.ModuleTask>();
        tasks.forEach((task) => {
            if (!task.name) {
                throw new SyntaxError("Missing task name!");
            }
            collection.set(task.name, this.parseModuleTask(module, task))
        });
        return new Models.ModuleTasks(collection);
    }

    private parseModuleInputParametersHash(
        module: Manifests.Module,
        inputs: Manifests.ModuleInputParameters
    ) {
        const collection = new Map<string, Models.ModuleInputParameter>();
        for (const name in inputs) {
            const input = inputs[name];
            if (!(input instanceof Object)) {
                throw new SyntaxError("Invalid input definition: not a object!");
            }
            collection.set(
                name,
                this.parseModuleInputParameter(
                    module,
                    _.merge({}, inputs[name], {name: name}),
                )
            );
        }
        return new Models.ModuleInputParameters(collection);
    }

    private parseModuleHelpersHash(
        module: Manifests.Module,
        helpers: Manifests.ModuleHelpers
    ) {
        const collection = new Map<string, Function>();
        for (const name in helpers) {
            let helper = helpers[name];
            if (typeof helper == "string") {
                helper = eval("(" + helper + ")");
            }
            if (!(helper instanceof Function)) {
                throw new SyntaxError("Invalid helper definition: not a function!");
            }
            collection.set(name, helper);
        }
        return new Models.ModuleHelpers(collection);
    }

    private parseModuleTasksHash(
        module: Manifests.Module,
        tasks: { [name: string]: Manifests.ModuleLifecycleActionTask }
    ) {
        const collection = new Map<string, Models.ModuleTask>();
        for (const name in tasks) {
            const task = tasks[name];
            if (!(task instanceof Object)) {
                throw new SyntaxError("Invalid task definition: not a object!");
            }
            collection.set(
                name,
                this.parseModuleTask(
                    module,
                    _.merge({}, tasks[name], {name: name}),
                )
            );
        }
        return new Models.ModuleTasks(collection);
    }

    private parseModuleTask(module: Manifests.Module, task: Manifests.ModuleTask) {
        if (!task.name) throw new SyntaxError("Missing task name!");
        //if (!task.executor) throw new SyntaxError(`Missing task '${task.name}' executor!`);
        return new Models.ModuleTask(
            task["name"],
            this.parseModuleTaskMetadata(module, task.metadata),
            task["configuration"]
        );
    }

    private parseModuleInputParameter(
        module: Manifests.Module,
        parameter: Manifests.ModuleInputParameter
    ) {
        if (!parameter.name) throw new SyntaxError("Missing input parameter name!");
        return new Models.ModuleInputParameter(
            parameter["name"],
            parameter["required"] == undefined ? true : parameter["required"],
            parameter["provided"] == undefined ? false : parameter["provided"],
            parameter["value"],
            parameter["prompt"] == undefined ? undefined : this.parseModuleInputParameterPrompt(
                module,
                parameter,
                parameter["prompt"]
            )
        );
    }

    private parseModuleTaskMetadata(
        module: Manifests.Module,
        metadata?: Manifests.ModuleTaskMetadata
    ) {
        return !metadata ?
            new Models.ModuleTaskMetadata() :
            new Models.ModuleTaskMetadata(
                metadata.enabled,
                metadata.executor,
                metadata["supports-virtual"],
                metadata.description
            );
    }

    private parseModuleInputParameterPrompt(
        module: Manifests.Module,
        parameter: Manifests.ModuleInputParameter,
        prompt: Manifests.ModuleInputParameterPrompt
    ) {
        if (!prompt) {
            return;
        }
        if (!prompt.message) {
            throw new SyntaxError("Missing prompt message!");
        }
        return new Models.ModuleInputParameterPrompt(
            prompt.message,
            prompt.type,
            prompt.choices,
            prompt.defaults
        );
    }
}
